# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import json
import torch
import torch.nn as nn
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from peft import LoraConfig, PeftModelForCausalLM
from transformers import LlamaModel, LlamaConfig, LlamaForCausalLM
from unitorch.models import (
    GenericModel,
    GenericOutputs,
    QuantizationConfig,
    QuantizationMixin,
)
from unitorch.models.quantization import quantize_model
from unitorch.models.peft import PeftModelForSequenceClassification, GenericPeftModel


class LlamaLoraForClassification(GenericPeftModel):
    prefix_keys_in_state_dict = {
        "^(?!peft_model\.base_model\.model\.).*": "peft_model.base_model."
    }

    def __init__(
        self,
        config_path: str,
        quant_config_path: Optional[str] = None,
        lora_r: Optional[int] = 16,
        lora_alpha: Optional[int] = 32,
        lora_dropout: Optional[float] = 0.05,
        fan_in_fan_out: Optional[bool] = True,
        target_modules: Optional[Union[List[str], str]] = ["q_proj", "v_proj"],
        num_classes: Optional[int] = 1,
        hidden_dropout_prob: Optional[float] = 0.1,
        gradient_checkpointing: Optional[bool] = False,
    ):
        super().__init__()
        self.config = LlamaConfig.from_json_file(config_path)
        self.config.gradient_checkpointing = gradient_checkpointing
        self.peft_config = LoraConfig(
            r=lora_r,
            lora_alpha=lora_alpha,
            lora_dropout=lora_dropout,
            fan_in_fan_out=fan_in_fan_out,
            target_modules=target_modules,
        )
        model = LlamaModel(self.config)
        if quant_config_path is not None:
            quant_config = QuantizationConfig.from_json_file(quant_config_path)
            ignore_modules = target_modules + ["lm_head"]
            model = quantize_model(model, quant_config, ignore_modules=ignore_modules)
        self.peft_model = PeftModelForSequenceClassification(model, self.peft_config)
        self.dropout = nn.Dropout(hidden_dropout_prob)
        self.classifier = nn.Linear(self.config.hidden_size, num_classes)
        self.init_weights()

    def forward(
        self,
        input_ids: torch.Tensor,
        attention_mask: Optional[torch.Tensor] = None,
        position_ids: Optional[torch.Tensor] = None,
    ):
        """
        Forward pass of the classification model.

        Args:
            input_ids (torch.Tensor): Input tensor of shape (batch_size, sequence_length).
            attention_mask (torch.Tensor, optional): Attention mask tensor of shape (batch_size, sequence_length). Defaults to None.
            position_ids (torch.Tensor, optional): Position IDs tensor of shape (batch_size, sequence_length). Defaults to None.

        Returns:
            torch Output logits.Tensor: tensor of shape (batch_size, num_classes).
        """
        outputs = self.peft_model(
            input_ids,
            attention_mask=attention_mask,
            position_ids=position_ids,
        )[0]
        pooled_output = outputs[:, -1]
        pooled_output = self.dropout(pooled_output)
        logits = self.classifier(pooled_output)
        return logits


class LlamaLoraForGeneration(GenericPeftModel):
    prefix_keys_in_state_dict = {
        "^(?!peft_model\.base_model\.model\.model\.)model\.": "peft_model.base_model.model.",
        "^lm_head.": "peft_model.base_model.model.",
    }

    def __init__(
        self,
        config_path: str,
        quant_config_path: Optional[str] = None,
        lora_r: Optional[int] = 16,
        lora_alpha: Optional[int] = 32,
        lora_dropout: Optional[float] = 0.05,
        fan_in_fan_out: Optional[bool] = True,
        target_modules: Optional[Union[List[str], str]] = ["q_proj", "v_proj"],
        gradient_checkpointing: Optional[bool] = False,
    ):
        super().__init__()
        self.config = LlamaConfig.from_json_file(config_path)
        self.config.gradient_checkpointing = gradient_checkpointing
        self.peft_config = LoraConfig(
            r=lora_r,
            lora_alpha=lora_alpha,
            lora_dropout=lora_dropout,
            fan_in_fan_out=fan_in_fan_out,
            target_modules=target_modules,
        )
        model = LlamaForCausalLM(self.config)
        if quant_config_path is not None:
            quant_config = QuantizationConfig.from_json_file(quant_config_path)
            ignore_modules = target_modules + ["lm_head"]
            model = quantize_model(model, quant_config, ignore_modules=ignore_modules)
        self.peft_model = PeftModelForCausalLM(model, self.peft_config)
        self.init_weights()

    def forward(
        self,
        input_ids: torch.Tensor,
        attention_mask: Optional[torch.Tensor] = None,
        position_ids: Optional[torch.Tensor] = None,
    ):
        """
        Forward pass of the generation model.

        Args:
            input_ids (torch.Tensor, optional): Input tensor of shape (batch_size, sequence_length). Defaults to None.
            attention_mask (torch.Tensor, optional): Attention mask tensor of shape (batch_size, sequence_length). Defaults to None.
            position_ids (torch.Tensor, optional): Position IDs tensor of shape (batch_size, sequence_length). Defaults to None.

        Returns:
            torch Output logits.Tensor: tensor of shape (batch_size, sequence_length, vocab_size).
        """
        outputs = self.peft_model(
            input_ids=input_ids,
            attention_mask=attention_mask,
            position_ids=position_ids,
            return_dict=True,
        )
        logits = outputs.logits
        return logits

    @torch.no_grad()
    def generate(
        self,
        input_ids: torch.Tensor,
        num_beams: Optional[int] = 5,
        decoder_start_token_id: Optional[int] = 1,
        decoder_end_token_id: Optional[Union[int, List[int]]] = 2,
        num_return_sequences: Optional[int] = 1,
        min_gen_seq_length: Optional[int] = 0,
        max_gen_seq_length: Optional[int] = 48,
        repetition_penalty: Optional[float] = 1.0,
        no_repeat_ngram_size: Optional[int] = 0,
        early_stopping: Optional[bool] = True,
        length_penalty: Optional[float] = 1.0,
        num_beam_groups: Optional[int] = 1,
        diversity_penalty: Optional[float] = 0.0,
        do_sample: Optional[bool] = False,
        temperature: Optional[float] = 1.0,
        top_k: Optional[int] = 50,
        top_p: Optional[float] = 1.0,
    ):
        """
        Generate text using the generation model.

        Args:
            input_ids: Input tensor of shape (batch_size, sequence_length).
            num_beams (int, optional): Number of beams for beam search. Defaults to 5.
            decoder_start_token_id (int, optional): The ID of the decoder start token. Defaults to 2.
            decoder_end_token_id (int or List[int], optional): The ID(s) of the decoder end token(s). Defaults to 2.
            num_return_sequences (int, optional): Number of generated sequences to return. Defaults to 1.
            min_gen_seq_length (int, optional): Minimum length of generated sequences. Defaults to 0.
            max_gen_seq_length (int, optional): Maximum length of generated sequences. Defaults to 48.
            repetition_penalty (float, optional): Penalty for repeated tokens. Defaults to 1.0.
            no_repeat_ngram_size (int, optional): Size of n-grams to avoid repeating. Defaults to 0.
            early_stopping (bool, optional): Whether to stop generation early. Defaults to True.
            length_penalty (float, optional): Penalty for longer sequences. Defaults to 1.0.
            num_beam_groups (int, optional): Number of beam groups for diverse beam search. Defaults to 1.
            diversity_penalty (float, optional): Penalty for diverse sequences in diverse beam search. Defaults to 0.0.
            do_sample (bool, optional): Whether to use sampling for generation. Defaults to False.
            temperature (float, optional): Sampling temperature. Defaults to 1.0.
            top_k (int, optional): Top-k value for sampling. Defaults to 50.
            top_p (float, optional): Top-p value for sampling. Defaults to 1.0.

        Returns:
            GenericOutputs: Generated sequences and their scores.
        """
        input_seq_length = input_ids.size(1)
        outputs = self.peft_model.generate(
            input_ids=input_ids,
            max_length=max_gen_seq_length + input_seq_length,
            min_length=min_gen_seq_length + input_seq_length,
            num_beams=num_beams,
            do_sample=do_sample,
            decoder_start_token_id=decoder_start_token_id,
            no_repeat_ngram_size=no_repeat_ngram_size,
            early_stopping=early_stopping,
            length_penalty=length_penalty,
            repetition_penalty=repetition_penalty,
            num_return_sequences=num_return_sequences,
            bos_token_id=decoder_start_token_id,
            eos_token_id=decoder_end_token_id,
            num_beam_groups=num_beam_groups,
            diversity_penalty=diversity_penalty,
            temperature=temperature,
            top_k=top_k,
            top_p=top_p,
            return_dict_in_generate=True,
            output_scores=True,
        )

        sequences = outputs.sequences.reshape(
            -1, num_return_sequences, outputs.sequences.size(-1)
        )
        outputs.sequences = torch.zeros(
            sequences.size(0), num_return_sequences, max_gen_seq_length
        ).to(device=sequences.device)
        outputs.sequences[:, :, : sequences.size(-1) - input_seq_length].copy_(
            sequences[:, :, input_seq_length : sequences.size(-1)]
        )

        if num_return_sequences == 1:
            outputs.sequences = outputs.sequences.reshape(-1, max_gen_seq_length)

        return GenericOutputs(
            sequences=outputs.sequences.long(),
            sequences_scores=outputs.sequences_scores,
        )



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import json
import numpy as np
import torch
import torch.nn.functional as F
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
import diffusers.schedulers as schedulers
from transformers import CLIPTextConfig, CLIPTextModel
from diffusers.schedulers import SchedulerMixin
from diffusers.models import AutoencoderKL, UNet2DConditionModel, UNetMotionModel
from diffusers.models.unets.unet_motion_model import MotionAdapter
from diffusers.image_processor import PipelineImageInput, VaeImageProcessor
from diffusers.pipelines.animatediff.pipeline_animatediff import (
    AnimateDiffPipeline as AnimateText2VideoPipeline,
    AnimateDiffPipelineOutput,
)
from unitorch.models import (
    GenericModel,
    GenericOutputs,
    QuantizationConfig,
    QuantizationMixin,
)


class GenericAnimateModel(GenericModel, QuantizationMixin):
    prefix_keys_in_state_dict = {
        # unet weights
        "^conv_in.*": "unet.",
        "^conv_norm_out.*": "unet.",
        "^conv_out.*": "unet.",
        "^time_embedding.*": "unet.",
        "^up_blocks.*": "unet.",
        "^mid_block.*": "unet.",
        "^down_blocks.*": "unet.",
        # text weights
        "^text_model.*": "text.",
        # vae weights
        "^encoder.*": "vae.",
        "^decoder.*": "vae.",
        "^post_quant_conv.*": "vae.",
        "^quant_conv.*": "vae.",
    }

    replace_keys_in_state_dict = {
        "\.query\.": ".to_q.",
        "\.key\.": ".to_k.",
        "\.value\.": ".to_v.",
        "\.proj_attn\.": ".to_out.0.",
    }

    def __init__(
        self,
        config_path: str,
        motion_config_path: str,
        text_config_path: str,
        vae_config_path: str,
        scheduler_config_path: str,
        quant_config_path: Optional[str] = None,
        image_size: Optional[int] = None,
        in_channels: Optional[int] = None,
        out_channels: Optional[int] = None,
        num_train_timesteps: Optional[int] = 1000,
        num_infer_timesteps: Optional[int] = 50,
        freeze_vae_encoder: Optional[bool] = True,
        freeze_text_encoder: Optional[bool] = True,
        seed: Optional[int] = 1123,
    ):
        super().__init__()
        self.seed = seed
        self.num_train_timesteps = num_train_timesteps
        self.num_infer_timesteps = num_infer_timesteps
        self.image_size = image_size

        config_dict = json.load(open(config_path))
        if image_size is not None:
            config_dict.update({"sample_size": image_size})
        if in_channels is not None:
            config_dict.update({"in_channels": in_channels})
        if out_channels is not None:
            config_dict.update({"out_channels": out_channels})
        self.unet = UNet2DConditionModel.from_config(config_dict)

        motion_config_dict = json.load(open(motion_config_path))
        self.motion = MotionAdapter.from_config(motion_config_dict)

        text_config = CLIPTextConfig.from_json_file(text_config_path)
        self.text = CLIPTextModel(text_config)

        vae_config_dict = json.load(open(vae_config_path))
        self.vae = AutoencoderKL.from_config(vae_config_dict)

        scheduler_config_dict = json.load(open(scheduler_config_path))
        scheduler_class_name = scheduler_config_dict.get("_class_name", "DDPMScheduler")
        assert hasattr(schedulers, scheduler_class_name)
        scheduler_class = getattr(schedulers, scheduler_class_name)
        assert issubclass(scheduler_class, SchedulerMixin)
        scheduler_config_dict["num_train_timesteps"] = num_train_timesteps
        self.scheduler = scheduler_class.from_config(scheduler_config_dict)

        self.vae_scale_factor = 2 ** (len(self.vae.config.block_out_channels) - 1)

        if freeze_vae_encoder:
            for param in self.vae.parameters():
                param.requires_grad = False

        if freeze_text_encoder:
            for param in self.text.parameters():
                param.requires_grad = False

        if quant_config_path is not None:
            self.quant_config = QuantizationConfig.from_json_file(quant_config_path)
            self.quantize(self.quant_config, ignore_modules=["lm_head", "unet", "vae"])

        self.scheduler.set_timesteps(num_inference_steps=self.num_infer_timesteps)


class AnimateForText2VideoGeneration(GenericAnimateModel):
    """
    AnimateForText2VideoGeneration is a class that represents a model for generating animated videos from text prompts.

    Args:
        config_path (str): The path to the configuration file.
        motion_config_path (str): The path to the motion configuration file.
        text_config_path (str): The path to the text configuration file.
        vae_config_path (str): The path to the VAE (Variational Autoencoder) configuration file.
        scheduler_config_path (str): The path to the scheduler configuration file.
        quant_config_path (str, optional): The path to the quantization configuration file. Defaults to None.
        image_size (int, optional): The size of the input images. Defaults to None.
        in_channels (int, optional): The number of input channels. Defaults to None.
        out_channels (int, optional): The number of output channels. Defaults to None.
        num_train_timesteps (int, optional): The number of training timesteps. Defaults to 1000.
        num_infer_timesteps (int, optional): The number of inference timesteps. Defaults to 50.
        freeze_vae_encoder (bool, optional): Whether to freeze the VAE encoder. Defaults to True.
        freeze_text_encoder (bool, optional): Whether to freeze the text encoder. Defaults to True.
        seed (int, optional): The random seed. Defaults to 1123.
    """

    def __init__(
        self,
        config_path: str,
        motion_config_path: str,
        text_config_path: str,
        vae_config_path: str,
        scheduler_config_path: str,
        quant_config_path: Optional[str] = None,
        image_size: Optional[int] = None,
        in_channels: Optional[int] = None,
        out_channels: Optional[int] = None,
        num_train_timesteps: Optional[int] = 1000,
        num_infer_timesteps: Optional[int] = 50,
        freeze_vae_encoder: Optional[bool] = True,
        freeze_text_encoder: Optional[bool] = True,
        seed: Optional[int] = 1123,
    ):
        """
        Initializes the AnimateModel object.

        Args:
            config_path (str): The path to the main configuration file.
            motion_config_path (str): The path to the motion configuration file.
            text_config_path (str): The path to the text configuration file.
            vae_config_path (str): The path to the VAE configuration file.
            scheduler_config_path (str): The path to the scheduler configuration file.
            quant_config_path (str, optional): The path to the quantization configuration file. Defaults to None.
            image_size (int, optional): The size of the input images. Defaults to None.
            in_channels (int, optional): The number of input channels. Defaults to None.
            out_channels (int, optional): The number of output channels. Defaults to None.
            num_train_timesteps (int, optional): The number of training timesteps. Defaults to 1000.
            num_infer_timesteps (int, optional): The number of inference timesteps. Defaults to 50.
            freeze_vae_encoder (bool, optional): Whether to freeze the VAE encoder. Defaults to True.
            freeze_text_encoder (bool, optional): Whether to freeze the text encoder. Defaults to True.
            seed (int, optional): The random seed. Defaults to 1123.
        """
        super().__init__(
            config_path=config_path,
            motion_config_path=motion_config_path,
            text_config_path=text_config_path,
            vae_config_path=vae_config_path,
            scheduler_config_path=scheduler_config_path,
            quant_config_path=quant_config_path,
            image_size=image_size,
            in_channels=in_channels,
            out_channels=out_channels,
            num_train_timesteps=num_train_timesteps,
            num_infer_timesteps=num_infer_timesteps,
            freeze_vae_encoder=freeze_vae_encoder,
            freeze_text_encoder=freeze_text_encoder,
            seed=seed,
        )

        self.pipeline = AnimateText2VideoPipeline(
            vae=self.vae,
            text_encoder=self.text,
            unet=self.unet,
            motion_adapter=self.motion,
            scheduler=self.scheduler,
            tokenizer=None,
        )
        self.pipeline.set_progress_bar_config(disable=True)

    def forward(
        self,
    ):
        """
        Performs the forward pass of the model.

        Raises:
            NotImplementedError: This method should be implemented in a subclass.
        """
        raise NotImplementedError

    def train(self, mode=True):
        """
        Trains the model.

        Args:
            mode (bool): If True, sets the model to training mode. If False, sets the model to evaluation mode.

        Returns:
            None
        """
        if not mode:
            self.pipeline.unet = UNetMotionModel.from_unet2d(self.unet, self.motion)
            self.pipeline.unet.to(device=self.pipeline.device)
            self.unet.eval()
            self.motion.eval()
            self.vae.eval()
            self.text.eval()
        else:
            self.unet.train()
            self.motion.train()
            self.vae.train()
            self.text.train()

    def to(self, *args, **kwargs):
        """
        Moves the model parameters and buffers to the specified device.

        Args:
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            The model with the parameters and buffers moved to the specified device.
        """
        self.pipeline.unet.to(*args, **kwargs)
        return super().to(*args, **kwargs)

    def generate(
        self,
        input_ids: torch.Tensor,
        negative_input_ids: torch.Tensor,
        attention_mask: Optional[torch.Tensor] = None,
        negative_attention_mask: Optional[torch.Tensor] = None,
        height: Optional[int] = 512,
        width: Optional[int] = 512,
        num_frames: Optional[int] = 16,
        guidance_scale: Optional[float] = 7.5,
    ):
        """
        Generates animated frames based on the given input_ids and negative_input_ids.

        Args:
            input_ids (torch.Tensor): The input tensor representing the prompt.
            negative_input_ids (torch.Tensor): The input tensor representing the negative prompt.
            attention_mask (Optional[torch.Tensor], optional): The attention mask tensor. Defaults to None.
            negative_attention_mask (Optional[torch.Tensor], optional): The attention mask tensor for the negative prompt. Defaults to None.
            height (Optional[int], optional): The height of the generated frames. Defaults to 512.
            width (Optional[int], optional): The width of the generated frames. Defaults to 512.
            num_frames (Optional[int], optional): The number of frames to generate. Defaults to 16.
            guidance_scale (Optional[float], optional): The scale factor for guidance. Defaults to 7.5.

        Returns:
            GenericOutputs: The generated frames.
        """
        prompt_embeds = self.text(
            input_ids,
            # attention_mask,
        )[0]
        negative_prompt_embeds = self.text(
            negative_input_ids,
            # negative_attention_mask,
        )[0]

        frames = self.pipeline(
            prompt=None,
            prompt_embeds=prompt_embeds,
            negative_prompt_embeds=negative_prompt_embeds,
            generator=torch.Generator(device=self.pipeline.device).manual_seed(
                self.seed
            ),
            height=height,
            width=width,
            num_frames=num_frames,
            num_inference_steps=self.num_infer_timesteps,
            guidance_scale=guidance_scale,
            output_type="pt",
        ).frames

        return GenericOutputs(frames=frames)


class AnimateForImage2VideoGeneration(GenericAnimateModel):
    pass



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import os
import warnings
import logging
import torch
import torch.nn as nn
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from peft import (
    PeftConfig,
    PeftType,
    PromptLearningConfig,
    PeftModelForSequenceClassification,
)
from unitorch.utils import replace
from unitorch.models import CheckpointMixin


@replace(PeftModelForSequenceClassification)
class PeftModelForSequenceClassification(PeftModelForSequenceClassification):
    def __init__(self, model, peft_config: PeftConfig, adapter_name="default"):
        super().__init__(model, peft_config, adapter_name)

    def forward(
        self,
        input_ids=None,
        attention_mask=None,
        inputs_embeds=None,
        output_attentions=None,
        output_hidden_states=None,
        return_dict=None,
        **kwargs,
    ):
        return_dict = (
            return_dict if return_dict is not None else self.config.use_return_dict
        )
        peft_config = self.active_peft_config
        if not isinstance(peft_config, PromptLearningConfig):
            return self.base_model(
                input_ids=input_ids,
                attention_mask=attention_mask,
                inputs_embeds=inputs_embeds,
                output_attentions=output_attentions,
                output_hidden_states=output_hidden_states,
                return_dict=return_dict,
                **kwargs,
            )

        labels = kwargs.pop("labels", None)

        batch_size = input_ids.shape[0]
        if attention_mask is not None:
            # concat prompt attention mask
            prefix_attention_mask = torch.ones(
                batch_size, peft_config.num_virtual_tokens
            ).to(self.device)
            attention_mask = torch.cat((prefix_attention_mask, attention_mask), dim=1)
        if kwargs.get("position_ids", None) is not None:
            warnings.warn(
                "Position ids are not supported for parameter efficient tuning. Ignoring position ids."
            )
            kwargs["position_ids"] = None
        kwargs.update(
            {
                "attention_mask": attention_mask,
                "labels": labels,
                "output_attentions": output_attentions,
                "output_hidden_states": output_hidden_states,
                "return_dict": return_dict,
            }
        )

        if peft_config.peft_type == PeftType.PREFIX_TUNING:
            return self._prefix_tuning_forward(input_ids=input_ids, **kwargs)
        else:
            if kwargs.get("token_type_ids", None) is not None:
                kwargs["token_type_ids"] = torch.cat(
                    (
                        torch.zeros(batch_size, peft_config.num_virtual_tokens).to(
                            self.device
                        ),
                        kwargs["token_type_ids"],
                    ),
                    dim=1,
                ).long()
            if inputs_embeds is None:
                inputs_embeds = self.word_embeddings(input_ids)
            prompts = self.get_prompt(batch_size=batch_size)
            prompts = prompts.to(inputs_embeds.dtype)
            inputs_embeds = torch.cat((prompts, inputs_embeds), dim=1)
            return self.base_model(inputs_embeds=inputs_embeds, **kwargs)


class PeftCheckpointMixin(CheckpointMixin):
    checkpoint_name = "pytorch_model.bin"

    modules_to_save_checkpoints = ["lora"]

    def save_checkpoint(
        self,
        ckpt_dir: str,
        weight_name: str = None,
        **kwargs,
    ):
        """
        Save the model's current state as a checkpoint.

        Args:
            ckpt_dir (str): Directory path to save the checkpoint.
            weight_name (str): Name of the weight file.

        Returns:
            None
        """
        if weight_name is None:
            weight_name = self.checkpoint_name
        state_dict = self.state_dict()
        state_dict = {
            k: v
            for k, v in state_dict.items()
            if any(m in k for m in self.modules_to_save_checkpoints)
        }
        weight_path = os.path.join(ckpt_dir, weight_name)
        torch.save(state_dict, weight_path)
        logging.info(f"{type(self).__name__} model save checkpoint to {weight_path}")

    def from_checkpoint(
        self,
        ckpt_dir: str,
        weight_name: Optional[str] = None,
        **kwargs,
    ):
        """
        Load model weights from a checkpoint.

        Args:
            ckpt_dir (str): Directory path of the checkpoint.
            weight_name (str): Name of the weight file.

        Returns:
            None
        """
        if weight_name is None:
            weight_name = self.checkpoint_name
        weight_path = os.path.join(ckpt_dir, weight_name)
        if not os.path.exists(weight_path):
            return
        _state_dict = self.state_dict()
        _state_dict = {
            k: v
            for k, v in _state_dict.items()
            if any(m in k for m in self.modules_to_save_checkpoints)
        }
        state_dict = torch.load(weight_path, map_location="cpu")
        assert all(
            k in state_dict.keys() and state_dict[k].shape == v.shape
            for k, v in _state_dict.items()
        )
        self.load_state_dict(state_dict, strict=False)
        logging.info(
            f"{type(self).__name__} model load weight from checkpoint {weight_path}"
        )


class GenericPeftModel(nn.Module, PeftCheckpointMixin):
    def __init__(self):
        super().__init__()
        pass

    def _init_weights(self, module):
        """
        Initialize the weights of the given module.

        Args:
            module (nn.Module): The module to initialize weights for.
        """
        if isinstance(module, (nn.Linear, nn.Embedding)):
            module.weight.data.normal_(mean=0.0, std=0.02)

        if isinstance(module, nn.LayerNorm):
            module.bias.data.zero_()
            module.weight.data.fill_(1.0)

        if isinstance(module, nn.Linear) and module.bias is not None:
            module.bias.data.zero_()

    def init_weights(self):
        """
        Initialize the weights of the model.
        """
        self.apply(self._init_weights)

    @property
    def dtype(self) -> torch.dtype:
        """
        Returns the data type of the model's parameters.

        Returns:
            torch.dtype: The data type of the model's parameters.
        """
        return next(self.parameters()).dtype

    @property
    def device(self):
        """
        Returns the device of the model's parameters.

        Returns:
            torch.device: The device of the model's parameters.
        """
        return next(self.parameters()).device


from unitorch.models.peft.modeling_bloom import (
    BloomLoraForClassification,
    BloomLoraForGeneration,
)
from unitorch.models.peft.modeling_llama import (
    LlamaLoraForClassification,
    LlamaLoraForGeneration,
)
from unitorch.models.peft.modeling_mistral import MistralLoraForGeneration



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

# pretrained infos

pretrained_deberta_infos = {
    "default-deberta": {
        "config": "https://huggingface.co/microsoft/deberta-base/resolve/main/config.json",
        "vocab": "https://huggingface.co/microsoft/deberta-base/resolve/main/vocab.json",
        "merge": "https://huggingface.co/microsoft/deberta-base/resolve/main/merges.txt",
    },
    "deberta-base": {
        "config": "https://huggingface.co/microsoft/deberta-base/resolve/main/config.json",
        "vocab": "https://huggingface.co/microsoft/deberta-base/resolve/main/vocab.json",
        "merge": "https://huggingface.co/microsoft/deberta-base/resolve/main/merges.txt",
        "weight": "https://huggingface.co/microsoft/deberta-base/resolve/main/pytorch_model.bin",
    },
    "deberta-large": {
        "config": "https://huggingface.co/microsoft/deberta-large/resolve/main/config.json",
        "vocab": "https://huggingface.co/microsoft/deberta-large/resolve/main/vocab.json",
        "merge": "https://huggingface.co/microsoft/deberta-large/resolve/main/merges.txt",
        "weight": "https://huggingface.co/microsoft/deberta-large/resolve/main/pytorch_model.bin",
    },
    "deberta-large-mnli": {
        "config": "https://huggingface.co/microsoft/deberta-large-mnli/resolve/main/config.json",
        "vocab": "https://huggingface.co/microsoft/deberta-large-mnli/resolve/main/vocab.json",
        "merge": "https://huggingface.co/microsoft/deberta-large-mnli/resolve/main/merges.txt",
        "weight": "https://huggingface.co/microsoft/deberta-large-mnli/resolve/main/pytorch_model.bin",
    },
    "deberta-xlarge": {
        "config": "https://huggingface.co/microsoft/deberta-xlarge/resolve/main/config.json",
        "vocab": "https://huggingface.co/microsoft/deberta-xlarge/resolve/main/vocab.json",
        "merge": "https://huggingface.co/microsoft/deberta-xlarge/resolve/main/merges.txt",
        "weight": "https://huggingface.co/microsoft/deberta-xlarge/resolve/main/pytorch_model.bin",
    },
}

pretrained_deberta_v2_infos = {
    "default-deberta-v2": {
        "config": "https://huggingface.co/microsoft/deberta-v2-xlarge/resolve/main/config.json",
        "vocab": "https://huggingface.co/microsoft/deberta-v2-xlarge/resolve/main/spm.model",
    },
    "deberta-v2-xlarge": {
        "config": "https://huggingface.co/microsoft/deberta-v2-xlarge/resolve/main/config.json",
        "vocab": "https://huggingface.co/microsoft/deberta-v2-xlarge/resolve/main/spm.model",
        "weight": "https://huggingface.co/microsoft/deberta-v2-xlarge/resolve/main/pytorch_model.bin",
    },
    "deberta-v2-xlarge": {
        "config": "https://huggingface.co/microsoft/deberta-v2-xlarge/resolve/main/config.json",
        "vocab": "https://huggingface.co/microsoft/deberta-v2-xlarge/resolve/main/spm.model",
        "weight": "https://huggingface.co/microsoft/deberta-v2-xlarge/resolve/main/pytorch_model.bin",
    },
    "deberta-v2-xxlarge": {
        "config": "https://huggingface.co/microsoft/deberta-v2-xxlarge/resolve/main/config.json",
        "vocab": "https://huggingface.co/microsoft/deberta-v2-xxlarge/resolve/main/spm.model",
        "weight": "https://huggingface.co/microsoft/deberta-v2-xxlarge/resolve/main/pytorch_model.bin",
    },
    "deberta-v2-xlarge-mnli": {
        "config": "https://huggingface.co/microsoft/deberta-v2-xlarge-mnli/resolve/main/config.json",
        "vocab": "https://huggingface.co/microsoft/deberta-v2-xlarge-mnli/resolve/main/spm.model",
        "weight": "https://huggingface.co/microsoft/deberta-v2-xlarge-mnli/resolve/main/pytorch_model.bin",
    },
    "deberta-v2-xxlarge-mnli": {
        "config": "https://huggingface.co/microsoft/deberta-v2-xxlarge-mnli/resolve/main/config.json",
        "vocab": "https://huggingface.co/microsoft/deberta-v2-xxlarge-mnli/resolve/main/spm.model",
        "weight": "https://huggingface.co/microsoft/deberta-v2-xxlarge-mnli/resolve/main/pytorch_model.bin",
    },
}

import unitorch.cli.models.deberta.modeling
import unitorch.cli.models.deberta.processing
import unitorch.cli.models.deberta.modeling_v2
import unitorch.cli.models.deberta.processing_v2
from unitorch.cli.models.deberta.modeling import DebertaForClassification
from unitorch.cli.models.deberta.processing import DebertaProcessor
from unitorch.cli.models.deberta.modeling_v2 import DebertaV2ForClassification
from unitorch.cli.models.deberta.processing_v2 import DebertaV2Processor



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import torch
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from torch.cuda.amp import autocast
from unitorch.utils import pop_value, nested_dict_value
from unitorch.models.deberta import (
    DebertaV2ForClassification as _DebertaV2ForClassification,
)
from unitorch.cli import (
    cached_path,
    add_default_section_for_init,
    add_default_section_for_function,
    register_model,
)
from unitorch.cli.models import ClassificationOutputs
from unitorch.cli.models.deberta import pretrained_deberta_v2_infos


@register_model("core/model/classification/deberta/v2")
class DebertaV2ForClassification(_DebertaV2ForClassification):
    """Deberta V2 model for classification tasks."""

    def __init__(
        self,
        config_path: str,
        num_classes: Optional[int] = 1,
        gradient_checkpointing: Optional[bool] = False,
    ):
        """
        Initialize the DebertaV2ForClassification model.

        Args:
            config_path (str): The path to the model configuration file.
            num_classes (int, optional): The number of classes for classification. Defaults to 1.
            gradient_checkpointing (bool, optional): Whether to use gradient checkpointing during training. Defaults to False.
        """
        super().__init__(
            config_path=config_path,
            num_classes=num_classes,
            gradient_checkpointing=gradient_checkpointing,
        )

    @classmethod
    @add_default_section_for_init("core/model/classification/deberta/v2")
    def from_core_configure(cls, config, **kwargs):
        """
        Create an instance of DebertaV2ForClassification from a core configuration.

        Args:
            config: The core configuration.
            **kwargs: Additional keyword arguments.

        Returns:
            DebertaV2ForClassification: An instance of DebertaV2ForClassification.
        """
        config.set_default_section("core/model/classification/deberta/v2")
        pretrained_name = config.getoption("pretrained_name", "default-deberta-v2")
        config_path = config.getoption("config_path", None)

        config_path = pop_value(
            config_path,
            nested_dict_value(pretrained_deberta_v2_infos, pretrained_name, "config"),
        )

        config_path = cached_path(config_path)
        gradient_checkpointing = config.getoption("gradient_checkpointing", False)
        num_classes = config.getoption("num_classes", 1)

        inst = cls(config_path, num_classes, gradient_checkpointing)
        pretrained_weight_path = config.getoption("pretrained_weight_path", None)
        weight_path = pop_value(
            pretrained_weight_path,
            nested_dict_value(pretrained_deberta_v2_infos, pretrained_name, "weight"),
            check_none=False,
        )
        if weight_path is not None:
            weight_path = cached_path(weight_path)
            inst.from_pretrained(weight_path)

        return inst

    @autocast()
    def forward(
        self,
        input_ids: torch.Tensor,
        attention_mask: Optional[torch.Tensor] = None,
        token_type_ids: Optional[torch.Tensor] = None,
        position_ids: Optional[torch.Tensor] = None,
    ):
        """
        Perform a forward pass on the DebertaV2ForClassification model.

        Args:
            input_ids (torch.Tensor): The input tensor containing the input IDs.
            attention_mask (torch.Tensor, optional): The attention mask tensor. Defaults to None.
            position_ids (torch.Tensor, optional): The position IDs tensor. Defaults to None.

        Returns:
            ClassificationOutputs: The output of the classification model.
        """
        outputs = super().forward(
            input_ids=input_ids,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            position_ids=position_ids,
        )
        return ClassificationOutputs(outputs=outputs)



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import re
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union


def matched_pretrained_names(
    pretrained_names: List[str],
    match_patterns: List[str],
    block_patterns: List[str] = [],
) -> List[str]:
    matched = [
        p for p in pretrained_names if any([re.match(m, p) for m in match_patterns])
    ]
    blacked = [p for p in matched if not any([re.match(b, p) for b in block_patterns])]
    return blacked


import unitorch.cli.webuis.image_utils
import unitorch.cli.webuis.animate
import unitorch.cli.webuis.blip2
import unitorch.cli.webuis.bloom
import unitorch.cli.webuis.stable
import unitorch.cli.webuis.controlnet
import unitorch.cli.webuis.stable_xl
import unitorch.cli.webuis.stable_xl_refiner
import unitorch.cli.webuis.controlnet_xl
import unitorch.cli.webuis.llama
import unitorch.cli.webuis.minigpt4
import unitorch.cli.webuis.mistral
import unitorch.cli.webuis.sam



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

from unitorch.models.mistral.modeling import (
    MistralForGeneration,
)
from unitorch.models.mistral.processing import MistralProcessor



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import os
import requests
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union
from datasets import config
from huggingface_hub.hf_api import (
    HfApi,
    DatasetInfo,
    hf_raise_for_status,
    quote,
    validate_hf_hub_args,
)
from unitorch.utils.decorators import replace


@replace(HfApi)
class HfApiV2(HfApi):
    @validate_hf_hub_args
    def dataset_info(
        self,
        repo_id: str,
        *,
        revision: Optional[str] = None,
        timeout: Optional[float] = None,
        files_metadata: Optional[bool] = False,
        token: Optional[Union[bool, str]] = None,
    ) -> DatasetInfo:
        """
        Get info on one specific dataset on huggingface.co.
        Dataset can be private if you pass an acceptable token.
        Args:
            repo_id (`str`):
                A namespace (user or an organization) and a repo name separated
                by a `/`.
            revision (`str`, *optional*):
                The revision of the dataset repository from which to get the
                information.
            token (`str`, *optional*):
                Deprecated in favor of `use_auth_token`. Will be removed in 0.12.0.
                An authentication token (See https://huggingface.co/settings/token)
            timeout (`float`, *optional*):
                Whether to set a timeout for the request to the Hub.
            files_metadata (`bool`, *optional*):
                Whether or not to retrieve metadata for files in the repository
                (size, LFS metadata, etc). Defaults to `False`.
            use_auth_token (`bool` or `str`, *optional*):
                Whether to use the `auth_token` provided from the
                `huggingface_hub` cli. If not logged in, a valid `auth_token`
                can be passed in as a string.
        Returns:
            [`hf_api.DatasetInfo`]: The dataset repository information.
        <Tip>
        Raises the following errors:
            - [`~utils.RepositoryNotFoundError`]
              If the repository to download from cannot be found. This may be because it doesn't exist,
              or because it is set to `private` and you do not have access.
            - [`~utils.RevisionNotFoundError`]
              If the revision to download from cannot be found.
        </Tip>
        """
        headers = self._build_hf_headers(token=token)
        params = {}
        if files_metadata:
            params["blobs"] = True
        hub_endpoint = os.environ.get("UNITORCH_HUB_ENDPOINT", None)
        if hub_endpoint is not None:
            try:
                path = (
                    f"{hub_endpoint}/api/datasets/{repo_id}"
                    if revision is None
                    else (
                        f"{hub_endpoint}/api/datasets/{repo_id}/revision/{quote(revision, safe='')}"
                    )
                )
                r = requests.get(path, headers=headers, params=params, timeout=timeout)
                config.HF_ENDPOINT = hub_endpoint
                config.HUB_DATASETS_URL = (
                    config.HF_ENDPOINT + "/datasets/{path}/resolve/{revision}/{name}"
                )
                config.HUB_DEFAULT_VERSION = "master"
            except:
                path = (
                    f"{self.endpoint}/api/datasets/{repo_id}"
                    if revision is None
                    else (
                        f"{self.endpoint}/api/datasets/{repo_id}/revision/{quote(revision, safe='')}"
                    )
                )
                r = requests.get(path, headers=headers, params=params, timeout=timeout)
        else:
            path = (
                f"{self.endpoint}/api/datasets/{repo_id}"
                if revision is None
                else (
                    f"{self.endpoint}/api/datasets/{repo_id}/revision/{quote(revision, safe='')}"
                )
            )
            r = requests.get(path, headers=headers, params=params, timeout=timeout)

        hf_raise_for_status(r)
        d = r.json()
        return DatasetInfo(**d)



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import re
from PIL import Image
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from unitorch.utils import pop_value, nested_dict_value
from unitorch.models.minigpt4 import (
    MiniGPT4Blip2LlamaProcessor as _MiniGPT4Blip2LlamaProcessor,
)
from unitorch.cli import (
    cached_path,
    add_default_section_for_init,
    add_default_section_for_function,
    register_process,
)
from unitorch.cli import WriterOutputs
from unitorch.cli.models import (
    TensorsInputs,
    GenerationOutputs,
    GenerationTargets,
)
from unitorch.cli.models.minigpt4 import pretrained_minigpt4_infos


class MiniGPT4Blip2LlamaProcessor(_MiniGPT4Blip2LlamaProcessor):
    """
    MiniGPT4Blip2LlamaProcessor is a class for processing inputs and outputs of the MiniGPT4 model with Blip2 and Llama.
    It inherits from the _MiniGPT4Blip2LlamaProcessor class.
    """

    def __init__(
        self,
        vocab_path: str,
        vision_config_path: str,
        max_prefix_seq_length: Optional[int] = 64,
        max_suffix_seq_length: Optional[int] = 128,
        max_gen_seq_length: Optional[int] = 128,
        weight_path: Optional[Union[str, List[str]]] = None,
        state_dict: Optional[Dict[str, Any]] = None,
        device: Optional[Union[str, int]] = "cpu",
    ):
        """
        Initializes a MiniGPT4Blip2LlamaProcessor instance.

        Args:
            vocab_path (str): The file path to the vocabulary.
            vision_config_path (str): The file path to the vision configuration.
            max_prefix_seq_length (int, optional): The maximum length of the prefix sequence. Defaults to 64.
            max_suffix_seq_length (int, optional): The maximum length of the suffix sequence. Defaults to 128.
            max_gen_seq_length (int, optional): The maximum length of the generated sequence. Defaults to 128.
        """
        super().__init__(
            vocab_file=vocab_path,
            vision_config_path=vision_config_path,
            max_prefix_seq_length=max_prefix_seq_length,
            max_suffix_seq_length=max_suffix_seq_length,
            max_gen_seq_length=max_gen_seq_length,
        )

    @classmethod
    @add_default_section_for_init("core/process/minigpt4")
    def from_core_configure(cls, config, **kwargs):
        """
        Creates a MiniGPT4Blip2LlamaProcessor instance from a core configuration.

        Args:
            config: The configuration object.
            **kwargs: Additional keyword arguments.

        Returns:
            dict: The dictionary containing the processor configuration.
        """
        config.set_default_section("core/process/minigpt4")
        pretrained_name = config.getoption("pretrained_name", "default-minigpt4")
        vocab_path = config.getoption("vocab_path", None)
        vocab_path = pop_value(
            vocab_path,
            nested_dict_value(pretrained_minigpt4_infos, pretrained_name, "vocab"),
        )
        vocab_path = cached_path(vocab_path)

        vision_config_path = config.getoption("vision_config_path", None)
        vision_config_path = pop_value(
            vision_config_path,
            nested_dict_value(
                pretrained_minigpt4_infos, pretrained_name, "vision_config"
            ),
        )
        vision_config_path = cached_path(vision_config_path)

        return {
            "vocab_path": vocab_path,
            "vision_config_path": vision_config_path,
        }

    @register_process("core/process/minigpt4/prompt")
    def _prompt(
        self,
        prefix_text: str,
        suffix_text: str,
        image: Union[Image.Image, str],
        max_prefix_seq_length: Optional[int] = None,
        max_suffix_seq_length: Optional[int] = None,
    ):
        """
        Processes inputs for the prompt mode.

        Args:
            prefix_text (str): The prefix text.
            suffix_text (str): The suffix text.
            image (Image.Image or str): The input image.
            max_prefix_seq_length (int, optional): The maximum length of the prefix sequence. Defaults to None.
            max_suffix_seq_length (int, optional): The maximum length of the suffix sequence. Defaults to None.

        Returns:
            TensorsInputs: The processed input tensors.
        """
        if isinstance(image, str):
            image = Image.open(image)
        outputs = super().prompt(
            prefix_text=prefix_text,
            suffix_text=suffix_text,
            image=image,
            max_prefix_seq_length=max_prefix_seq_length,
            max_suffix_seq_length=max_suffix_seq_length,
        )
        return TensorsInputs(
            prefix_input_ids=outputs.prefix_input_ids,
            suffix_input_ids=outputs.suffix_input_ids,
            pixel_values=outputs.pixel_values,
        )

    @register_process("core/process/minigpt4/generation/inputs")
    def _generation_inputs(
        self,
        prefix_text: str,
        suffix_text: str,
        image: Union[Image.Image, str],
        max_prefix_seq_length: Optional[int] = None,
        max_suffix_seq_length: Optional[int] = None,
    ):
        """
        Processes inputs for the generation mode.

        Args:
            prefix_text (str): The prefix text.
            suffix_text (str): The suffix text.
            image (Image.Image or str): The input image.
            max_prefix_seq_length (int, optional): The maximum length of the prefix sequence. Defaults to None.
            max_suffix_seq_length (int, optional): The maximum length of the suffix sequence. Defaults to None.

        Returns:
            TensorsInputs: The processed input tensors.
        """
        if isinstance(image, str):
            image = Image.open(image)
        outputs = super().generation_inputs(
            prefix_text=prefix_text,
            suffix_text=suffix_text,
            image=image,
            max_prefix_seq_length=max_prefix_seq_length,
            max_suffix_seq_length=max_suffix_seq_length,
        )
        return TensorsInputs(
            prefix_input_ids=outputs.prefix_input_ids,
            suffix_input_ids=outputs.suffix_input_ids,
            pixel_values=outputs.pixel_values,
        )

    @register_process("core/process/minigpt4/generation/labels")
    def _generation_labels(
        self,
        text: str,
        max_gen_seq_length: Optional[int] = None,
    ):
        """
        Processes labels for the generation mode.

        Args:
            text (str): The input text.
            max_gen_seq_length (int, optional): The maximum length of the generated sequence. Defaults to None.

        Returns:
            GenerationTargets: The processed generation targets.
        """
        outputs = super().generation_labels(
            text=text,
            max_gen_seq_length=max_gen_seq_length,
        )
        return GenerationTargets(
            refs=outputs.input_ids,
            masks=outputs.attention_mask,
        )

    @register_process("core/process/minigpt4/generation")
    def _generation(
        self,
        prefix_text: str,
        suffix_text: str,
        text_pair: str,
        image: Union[Image.Image, str],
        max_prefix_seq_length: Optional[int] = None,
        max_suffix_seq_length: Optional[int] = None,
        max_gen_seq_length: Optional[int] = None,
    ):
        """
        Processes inputs and labels for the generation mode.

        Args:
            prefix_text (str): The prefix text.
            suffix_text (str): The suffix text.
            text_pair (str): The text pair.
            image (Image.Image or str): The input image.
            max_prefix_seq_length (int, optional): The maximum length of the prefix sequence. Defaults to None.
            max_suffix_seq_length (int, optional): The maximum length of the suffix sequence. Defaults to None.
            max_gen_seq_length (int, optional): The maximum length of the generated sequence. Defaults to None.

        Returns:
            TensorsInputs: The processed input tensors.
            GenerationTargets: The processed generation targets.
        """
        if isinstance(image, str):
            image = Image.open(image)
        outputs = super().generation(
            prefix_text=prefix_text,
            suffix_text=suffix_text,
            text_pair=text_pair,
            image=image,
            max_prefix_seq_length=max_prefix_seq_length,
            max_suffix_seq_length=max_suffix_seq_length,
            max_gen_seq_length=max_gen_seq_length,
        )
        return TensorsInputs(
            prefix_input_ids=outputs.prefix_input_ids,
            prefix_attention_mask=outputs.prefix_attention_mask,
            suffix_input_ids=outputs.suffix_input_ids,
            suffix_attention_mask=outputs.suffix_attention_mask,
            decoder_input_ids=outputs.input_ids_pair,
            decoder_attention_mask=outputs.attention_mask_pair,
            pixel_values=outputs.pixel_values,
        ), GenerationTargets(
            refs=outputs.input_ids_label,
            masks=outputs.attention_mask_label,
        )

    @register_process("core/postprocess/minigpt4/detokenize")
    def _detokenize(
        self,
        outputs: GenerationOutputs,
    ):
        """
        Detokenizes the generation outputs.

        Args:
            outputs (GenerationOutputs): The generation outputs.

        Returns:
            WriterOutputs: The processed writer outputs.
        """
        results = outputs.to_pandas()
        assert results.shape[0] == 0 or results.shape[0] == outputs.sequences.shape[0]

        decoded = super().detokenize(sequences=outputs.sequences)
        cleanup_string = lambda text: re.sub(r"###|\n", " ", text)
        if isinstance(decoded[0], list):
            decoded = [list(map(cleanup_string, sequence)) for sequence in decoded]
        elif isinstance(decoded[0], str):
            decoded = list(map(cleanup_string, decoded))
        else:
            raise ValueError(
                f"Unsupported type for minigpt4 detokenize: {type(decoded[0])}"
            )
        results["decoded"] = decoded
        return WriterOutputs(results)



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import json
from transformers.utils.quantization_config import BitsAndBytesConfig
from transformers.utils.bitsandbytes import (
    replace_with_bnb_linear,
    set_module_quantized_tensor_to_device,
)
from unitorch.utils import is_bitsandbytes_available


def quantize_model(model, config, ignore_modules):
    assert is_bitsandbytes_available(), "Please install bitsandbytes first."
    params = {n: p.clone() for n, p in model.named_parameters()}
    buffers = {n: b.clone() for n, b in model.named_buffers()}
    model = replace_with_bnb_linear(
        model,
        modules_to_not_convert=ignore_modules,
        quantization_config=config,
    )
    if ignore_modules is None:
        ignore_modules = ["lm_head"]
    for name, value in {**params, **buffers}.items():
        if any(m in name for m in ignore_modules):
            continue
        set_module_quantized_tensor_to_device(model, name, value.device, value)
    return model


class QuantizationConfig(BitsAndBytesConfig):
    @classmethod
    def from_json_file(cls, json_file: str):
        with open(json_file, "r", encoding="utf-8") as f:
            config = json.load(f)
        return cls(**config)


class QuantizationMixin:
    def quantize(self, config: QuantizationConfig, ignore_modules=None):
        return quantize_model(self, config, ignore_modules)



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import os
import torch
import json
import numpy as np
from PIL import Image
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from transformers import BertTokenizer, CLIPTokenizer
from torchvision.transforms import (
    Resize,
    CenterCrop,
    RandomCrop,
    ToTensor,
    Normalize,
    Compose,
    RandomHorizontalFlip,
)
from diffusers.image_processor import VaeImageProcessor
from diffusers.pipelines.blip_diffusion.blip_image_processing import BlipImageProcessor
from unitorch.models import (
    HfImageClassificationProcessor,
    HfTextClassificationProcessor,
    GenericOutputs,
)


class Blip2Processor(HfImageClassificationProcessor, HfTextClassificationProcessor):
    def __init__(
        self,
        vocab_path: str,
        merge_path: str,
        vision_config_path: str,
        vae_config_path: Optional[str] = None,
        max_seq_length: Optional[int] = 77,
        position_start_id: Optional[int] = 0,
        pad_token: Optional[str] = "<|endoftext|>",
        image_size: Optional[int] = 512,
        center_crop: Optional[bool] = False,
    ):
        tokenizer = CLIPTokenizer(
            vocab_file=vocab_path,
            merges_file=merge_path,
        )

        tokenizer.cls_token = tokenizer.bos_token
        tokenizer.sep_token = tokenizer.eos_token
        tokenizer.pad_token = pad_token

        HfTextClassificationProcessor.__init__(
            self,
            tokenizer=tokenizer,
            max_seq_length=max_seq_length,
            position_start_id=position_start_id,
        )

        vision_processor = BlipImageProcessor.from_json_file(vision_config_path)

        HfImageClassificationProcessor.__init__(
            self,
            vision_processor=vision_processor,
        )

        self.refer_text_processor = HfTextClassificationProcessor(
            tokenizer=BertTokenizer.from_pretrained("bert-base-uncased"),
            max_seq_length=max_seq_length,
        )

        if vae_config_path is not None:
            vae_config_dict = json.load(open(vae_config_path))
            vae_scale_factor = 2 ** (
                len(vae_config_dict.get("block_out_channels", [])) - 1
            )
            self.vae_image_processor = VaeImageProcessor(
                vae_scale_factor=vae_scale_factor
            )
        else:
            self.vae_image_processor = None

        self.vae_condition_image_processor = VaeImageProcessor(
            vae_scale_factor=vae_scale_factor,
            do_convert_rgb=True,
            do_normalize=False,
        )

    def text2image(
        self,
    ):
        raise NotImplementedError

    def text2image_inputs(
        self,
        prompt: str,
        refer_prompt: str,
        refer_image: Union[Image.Image, str],
        negative_prompt: Optional[str] = "",
        max_seq_length: Optional[int] = None,
        max_qtokens: Optional[int] = 16,
    ):
        if isinstance(refer_image, str):
            refer_image = Image.open(refer_image).convert("RGB")

        max_seq_length = max_seq_length or self.max_seq_length
        prompt_outputs = HfTextClassificationProcessor.classification(
            self,
            text=prompt,
            max_seq_length=max_seq_length - max_qtokens,
        )
        refer_prompt_outputs = self.refer_text_processor.classification(
            text=refer_prompt,
            max_seq_length=max_seq_length,
        )
        refer_image_outputs = HfImageClassificationProcessor.classification(
            self,
            image=refer_image,
        )
        negative_prompt_outputs = HfTextClassificationProcessor.classification(
            self,
            text=negative_prompt,
            max_seq_length=max_seq_length,
        )
        return GenericOutputs(
            input_ids=prompt_outputs.input_ids,
            attention_mask=prompt_outputs.attention_mask,
            negative_input_ids=negative_prompt_outputs.input_ids,
            negative_attention_mask=negative_prompt_outputs.attention_mask,
            refer_input_ids=refer_prompt_outputs.input_ids,
            refer_attention_mask=refer_prompt_outputs.attention_mask,
            refer_pixel_values=refer_image_outputs.pixel_values,
        )

    def text2image_controlnet_inputs(
        self,
        prompt: str,
        condition_image: Union[Image.Image, str],
        refer_prompt: str,
        refer_image: Union[Image.Image, str],
        negative_prompt: Optional[str] = "",
        max_seq_length: Optional[int] = None,
        max_qtokens: Optional[int] = 16,
    ):
        if isinstance(condition_image, str):
            condition_image = Image.open(condition_image)

        text2image_inputs = self.text2image_inputs(
            prompt=prompt,
            refer_prompt=refer_prompt,
            refer_image=refer_image,
            negative_prompt=negative_prompt,
            max_seq_length=max_seq_length,
            max_qtokens=max_qtokens,
        )

        condition_pixel_values = self.vae_condition_image_processor.preprocess(
            condition_image
        )[0]

        return GenericOutputs(
            input_ids=text2image_inputs.input_ids,
            attention_mask=text2image_inputs.attention_mask,
            negative_input_ids=text2image_inputs.negative_input_ids,
            negative_attention_mask=text2image_inputs.negative_attention_mask,
            refer_input_ids=text2image_inputs.refer_input_ids,
            refer_attention_mask=text2image_inputs.refer_attention_mask,
            refer_pixel_values=text2image_inputs.refer_pixel_values,
            condition_pixel_values=condition_pixel_values,
        )



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

from itertools import repeat
import collections.abc


def _ntuple(n):
    def parse(x):
        if isinstance(x, collections.abc.Iterable) and not isinstance(x, str):
            return x
        return tuple(repeat(x, n))

    return parse


to_1tuple = _ntuple(1)
to_2tuple = _ntuple(2)
to_3tuple = _ntuple(3)
to_4tuple = _ntuple(4)
to_ntuple = _ntuple


def make_divisible(
    v,
    divisor=8,
    min_value=None,
    round_limit=0.9,
):
    """
    Rounds a value to the nearest multiple of a divisor while ensuring it satisfies certain constraints.

    Args:
        v (int): The input value to be made divisible.
        divisor (int, optional): The divisor value. Defaults to 8.
        min_value (int, optional): The minimum value allowed. Defaults to None.
        round_limit (float, optional): The rounding limit, expressed as a fraction of the original value.
            Specifies the maximum allowed decrease in the rounded value compared to the original value.
            Defaults to 0.9, indicating a limit of 10%.

    Returns:
        int: The rounded value that is divisible by the given divisor and satisfies the constraints.

    Notes:
        - If min_value is not provided (None), it defaults to the value of the divisor.
        - The function rounds the input value (v) to the nearest multiple of the divisor.
        - If the rounded value is less than round_limit times the original value, the function increments
          the rounded value by the divisor to ensure the round down does not go down by more than the limit.

    Example:
        >>> make_divisible(25)
        24
        >>> make_divisible(32, divisor=10)
        30
        >>> make_divisible(10, divisor=3, min_value=6)
        12
    """
    min_value = min_value or divisor
    new_v = max(min_value, int(v + divisor / 2) // divisor * divisor)
    # Make sure that round down does not go down by more than 10%.
    if new_v < round_limit * v:
        new_v += divisor
    return new_v



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

from unitorch.cli.pipelines.stable.text2image import (
    StableForText2ImageGenerationPipeline,
)
from unitorch.cli.pipelines.stable.image2image import (
    StableForImage2ImageGenerationPipeline,
)
from unitorch.cli.pipelines.stable.inpainting import StableForImageInpaintingPipeline
from unitorch.cli.pipelines.stable.resolution import StableForImageResolutionPipeline



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import os
import sys
import fire
import logging
import importlib
import unitorch.cli
from transformers.utils import is_remote_url
from unitorch.cli import CoreConfigureParser
from unitorch.cli import (
    import_library,
    cached_path,
    set_global_config,
    registered_task,
    registered_script,
    init_registered_module,
)


@fire.decorators.SetParseFn(str)
def eval(config_path_or_dir: str, **kwargs):
    config_file = kwargs.pop("config_file", "config.ini")

    if os.path.isdir(config_path_or_dir):
        config_path = os.path.join(config_path_or_dir, config_file)
        sys.path.insert(0, config_path_or_dir)
        for f in os.listdir(config_path_or_dir):
            fpath = os.path.normpath(os.path.join(config_path_or_dir, f))
            if (
                not f.startswith("_")
                and not f.startswith(".")
                and (f.endswith(".py") or os.path.isdir(fpath))
            ):
                fname = f[:-3] if f.endswith(".py") else f
                module = importlib.import_module(f"{fname}")
    else:
        config_path = cached_path(config_path_or_dir)

    params = []
    for k, v in kwargs.items():
        if k.count("@") > 0:
            k0 = k.split("@")[0]
            k1 = "@".join(k.split("@")[1:])
        else:
            k0 = "core/cli"
            k1 = k
        params.append((k0, k1, v))

    if config_path is not None:
        config = CoreConfigureParser(config_path, params=params)
    else:
        config = CoreConfigureParser(params=params)

    set_global_config(config)

    depends_libraries = config.getdefault("core/cli", "depends_libraries", None)

    if depends_libraries:
        for library in depends_libraries:
            import_library(library)

    task_name = config.getdefault("core/cli", "task_name", None)
    assert task_name is not None and task_name in registered_task
    cli_task = init_registered_module(task_name, config, registered_task)

    cli_task.eval()

    os._exit(0)


def cli_main():
    fire.Fire(eval)



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import sys
import logging
import json
import numpy as np
from unitorch import (
    is_deepspeed_available,
    is_accelerate_available,
    is_megatron_available,
)


import unitorch.cli.tasks.supervised
from unitorch.cli.tasks.supervised import SupervisedTask

if is_deepspeed_available():
    import unitorch.cli.tasks.deepspeed
    from unitorch.cli.tasks.deepspeed import DeepspeedTask



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import torch
from PIL import Image
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union
import unitorch.utils.palette


def make_grid(
    images: List[Image.Image], rows: int, cols: int, resize: Optional[List[int]] = None
) -> Image.Image:
    """
    Combines a list of images into a grid layout.

    Args:
        images (List[Image.Image]): List of PIL Images to combine into a grid.
        rows (int): Number of rows in the grid.
        cols (int): Number of columns in the grid.
        resize (Optional[List[int]], optional): Size to resize the images. Defaults to None.

    Returns:
        Image.Image: PIL Image object representing the grid layout.

    Raises:
        AssertionError: If the number of images is not equal to the product of rows and cols.
    """
    assert (
        len(images) == rows * cols
    ), "Number of images must be equal to the product of rows and cols."

    if resize is not None:
        w, h = resize
    else:
        w, h = images[0].size

    grid = Image.new("RGB", (w * cols, h * rows))
    for i, image in enumerate(images):
        grid.paste(image, box=(w * (i % cols), h * (i // cols)))

    return grid


def resize_shortest_edge(
    image: Image.Image,
    short_size: List[int],
    max_size: int,
):
    """
    Resize the image to the given short size and maximum size.

    Args:
        image (Image.Image): Input image.
        short_size (List[int]): Shortest edge size.
        max_size (int): Maximum size.

    Returns:
        Image.Image: Resized image.
    """
    w, h = image.size
    size = min(w, h)
    scale = min(short_size[0] / size, short_size[1] / size)
    if scale * size > max_size:
        scale = max_size / size
    new_w, new_h = int(w * scale), int(h * scale)
    return image.resize((new_w, new_h), Image.BILINEAR)


def image_list_to_tensor(
    images: List[torch.Tensor],
    size_divisibility: int = 0,
    pad_value: float = 0.0,
    padding_constraints: Optional[Dict[str, int]] = None,
) -> torch.Tensor:
    """
    Convert a list of images to a tensor.

    Args:
        images (List[torch.Tensor]): List of images to convert to a tensor.
        size_divisibility (int, optional): Size divisibility. Defaults to 0.
        pad_value (float, optional): Padding value. Defaults to 0.0.
        padding_constraints (Optional[Dict[str, int]], optional): Padding constraints. Defaults to None.

    Returns:
        torch.Tensor: Tensor representing the images.
    """
    if padding_constraints is None:
        padding_constraints = {"top": 0, "bottom": 0, "left": 0, "right": 0}
    max_size = tuple(max(s) for s in zip(*[img.shape for img in images]))
    if size_divisibility > 0:
        stride = size_divisibility
        max_size = [(s + stride - 1) // stride * stride for s in max_size]
    batch_shape = (len(images),) + max_size
    batched_imgs = images[0].new_full(batch_shape, pad_value)
    for img, pad in zip(images, batched_imgs):
        pad[
            : img.shape[0],
            : img.shape[1],
            : img.shape[2],
        ] = img
    return batched_imgs


def numpy_to_pil(images):
    """
    Convert a numpy image or a batch of images to a PIL image.
    """
    if images.ndim == 3:
        # single image
        images = (images * 255).round().astype("uint8")
        if images.shape[-1] == 1:
            # special case for grayscale (single channel) images
            return Image.fromarray(images.squeeze(), mode="L")
        else:
            return Image.fromarray(images)

    images = (images * 255).round().astype("uint8")
    if images.shape[-1] == 1:
        # special case for grayscale (single channel) images
        pil_images = [Image.fromarray(image.squeeze(), mode="L") for image in images]
    else:
        pil_images = [Image.fromarray(image) for image in images]

    return pil_images



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import torch
from dataclasses import dataclass
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from unitorch.cli import WriterMixin, WriterOutputs
from unitorch.cli.models.modeling_utils import TensorsOutputs, TensorsTargets


@dataclass
class RankingOutputs(TensorsOutputs, WriterMixin):
    outputs: torch.Tensor


@dataclass
class RankingTargets(TensorsTargets):
    targets: torch.Tensor
    masks: Optional[torch.Tensor] = torch.empty(0)



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import torch
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from torch.cuda.amp import autocast
from transformers.utils import is_remote_url
from unitorch.utils import pop_value, nested_dict_value
from unitorch.models.minigpt4 import (
    MiniGPT4Blip2LlamaForGeneration as _MiniGPT4Blip2LlamaForGeneration,
)
from unitorch.cli import (
    cached_path,
    add_default_section_for_init,
    add_default_section_for_function,
    register_model,
)
from unitorch.cli.models import generation_model_decorator
from unitorch.cli.models import ClassificationOutputs, GenerationOutputs, LossOutputs
from unitorch.cli.models.minigpt4 import pretrained_minigpt4_infos


@register_model("core/model/generation/minigpt4", generation_model_decorator)
class MiniGPT4Blip2LlamaForGeneration(_MiniGPT4Blip2LlamaForGeneration):
    """
    MiniGPT4Blip2LlamaForGeneration is a class for generating sequences using the MiniGPT4 model with Blip2 and Llama.
    It inherits from the _MiniGPT4Blip2LlamaForGeneration class.
    """

    def __init__(
        self,
        blip2_config_path: str,
        llama_config_path: str,
        quant_config_path: Optional[str] = None,
        pad_token_id: Optional[int] = 0,
        freeze_vision_model: Optional[bool] = True,
        freeze_qformer_model: Optional[bool] = True,
        freeze_llama_model: Optional[bool] = True,
        gradient_checkpointing: Optional[bool] = False,
    ):
        """
        Initializes a MiniGPT4Blip2LlamaForGeneration instance.

        Args:
            blip2_config_path (str): The file path to the Blip2 configuration.
            llama_config_path (str): The file path to the Llama configuration.
            pad_token_id (int, optional): The ID of the padding token. Defaults to 0.
            freeze_vision_model (bool, optional): Whether to freeze the vision model. Defaults to True.
            freeze_qformer_model (bool, optional): Whether to freeze the query transformer model. Defaults to True.
            freeze_llama_model (bool, optional): Whether to freeze the Llama model. Defaults to True.
            gradient_checkpointing (bool, optional): Whether to use gradient checkpointing. Defaults to False.
        """
        super().__init__(
            blip2_config_path=blip2_config_path,
            llama_config_path=llama_config_path,
            quant_config_path=quant_config_path,
            pad_token_id=pad_token_id,
            freeze_vision_model=freeze_vision_model,
            freeze_qformer_model=freeze_qformer_model,
            freeze_llama_model=freeze_llama_model,
            gradient_checkpointing=gradient_checkpointing,
        )

    @classmethod
    @add_default_section_for_init("core/model/generation/minigpt4")
    def from_core_configure(cls, config, **kwargs):
        """
        Creates a MiniGPT4Blip2LlamaForGeneration instance from a core configuration.

        Args:
            config: The configuration object.
            **kwargs: Additional keyword arguments.

        Returns:
            MiniGPT4Blip2LlamaForGeneration: The created instance.
        """
        config.set_default_section("core/model/generation/minigpt4")
        pretrained_name = config.getoption("pretrained_name", "default-minigpt4")

        blip2_config_path = config.getoption("blip2_config_path", None)
        blip2_config_path = pop_value(
            blip2_config_path,
            nested_dict_value(
                pretrained_minigpt4_infos, pretrained_name, "blip2_config_path"
            ),
        )
        blip2_config_path = cached_path(blip2_config_path)

        llama_config_path = config.getoption("llama_config_path", None)
        llama_config_path = pop_value(
            llama_config_path,
            nested_dict_value(
                pretrained_minigpt4_infos, pretrained_name, "llama_config_path"
            ),
        )
        llama_config_path = cached_path(llama_config_path)

        quant_config_path = config.getoption("quant_config_path", None)
        if quant_config_path is not None:
            quant_config_path = cached_path(quant_config_path)

        freeze_vision_model = config.getoption("freeze_vision_model", True)
        freeze_qformer_model = config.getoption("freeze_qformer_model", True)
        freeze_llama_model = config.getoption("freeze_llama_model", True)
        gradient_checkpointing = config.getoption("gradient_checkpointing", False)

        inst = cls(
            blip2_config_path,
            llama_config_path,
            quant_config_path=quant_config_path,
            freeze_vision_model=freeze_vision_model,
            freeze_qformer_model=freeze_qformer_model,
            freeze_llama_model=freeze_llama_model,
            gradient_checkpointing=gradient_checkpointing,
        )
        pretrained_weight_path = config.getoption("pretrained_weight_path", None)
        weight_path = pop_value(
            pretrained_weight_path,
            nested_dict_value(pretrained_minigpt4_infos, pretrained_name, "weight"),
            check_none=False,
        )

        if weight_path is not None:
            inst.from_pretrained(
                weight_path=weight_path,
            )

        return inst

    def forward(
        self,
        pixel_values: torch.Tensor,
        prefix_input_ids: torch.Tensor,
        suffix_input_ids: torch.Tensor,
        decoder_input_ids: torch.Tensor,
        prefix_attention_mask: Optional[torch.Tensor] = None,
        suffix_attention_mask: Optional[torch.Tensor] = None,
        decoder_attention_mask: Optional[torch.Tensor] = None,
    ):
        """
        Performs a forward pass through the model.

        Args:
            pixel_values (torch.Tensor): The pixel values.
            prefix_input_ids (torch.Tensor): The input IDs for the prefix tokens.
            suffix_input_ids (torch.Tensor): The input IDs for the suffix tokens.
            decoder_input_ids (torch.Tensor): The input IDs for the decoder tokens.
            prefix_attention_mask (torch.Tensor, optional): The attention mask for the prefix tokens.
                Defaults to None.
            suffix_attention_mask (torch.Tensor, optional): The attention mask for the suffix tokens.
                Defaults to None.
            decoder_attention_mask (torch.Tensor, optional): The attention mask for the decoder tokens.
                Defaults to None.

        Returns:
            GenerationOutputs: The generation outputs.
        """
        outputs = super().forward(
            pixel_values=pixel_values,
            prefix_input_ids=prefix_input_ids,
            suffix_input_ids=suffix_input_ids,
            decoder_input_ids=decoder_input_ids,
            prefix_attention_mask=prefix_attention_mask,
            suffix_attention_mask=suffix_attention_mask,
            decoder_attention_mask=decoder_attention_mask,
        )
        return GenerationOutputs(sequences=outputs)

    @add_default_section_for_function("core/model/generation/minigpt4")
    @torch.no_grad()
    def generate(
        self,
        pixel_values: torch.Tensor,
        prefix_input_ids: torch.Tensor,
        suffix_input_ids: torch.Tensor,
        num_beams: Optional[int] = 5,
        decoder_start_token_id: Optional[int] = 1,
        decoder_end_token_id: Optional[Union[int, List[int]]] = 2,
        num_return_sequences: Optional[int] = 1,
        min_gen_seq_length: Optional[int] = 0,
        max_gen_seq_length: Optional[int] = 48,
        repetition_penalty: Optional[float] = 1.0,
        no_repeat_ngram_size: Optional[int] = 0,
        early_stopping: Optional[bool] = True,
        length_penalty: Optional[float] = 1.0,
        num_beam_groups: Optional[int] = 1,
        diversity_penalty: Optional[float] = 0.0,
        do_sample: Optional[bool] = False,
        temperature: Optional[float] = 1.0,
        top_k: Optional[int] = 50,
        top_p: Optional[float] = 1.0,
    ):
        """
        Generates sequences using the model.

        Args:
            pixel_values (torch.Tensor): The pixel values.
            prefix_input_ids (torch.Tensor): The input IDs for the prefix tokens.
            suffix_input_ids (torch.Tensor): The input IDs for the suffix tokens.
            num_beams (int, optional): The number of beams for beam search. Defaults to 5.
            decoder_start_token_id (int, optional): The ID of the decoder start token. Defaults to 1.
            decoder_end_token_id (int or List[int], optional): The ID(s) of the decoder end token(s).
                Defaults to 2.
            num_return_sequences (int, optional): The number of sequences to return. Defaults to 1.
            min_gen_seq_length (int, optional): The minimum generated sequence length. Defaults to 0.
            max_gen_seq_length (int, optional): The maximum generated sequence length. Defaults to 48.
            repetition_penalty (float, optional): The repetition penalty. Defaults to 1.0.
            no_repeat_ngram_size (int, optional): The size of the n-grams to avoid repeating. Defaults to 0.
            early_stopping (bool, optional): Whether to perform early stopping. Defaults to True.
            length_penalty (float, optional): The length penalty. Defaults to 1.0.
            num_beam_groups (int, optional): The number of beam groups for diverse beam search.
                Defaults to 1.
            diversity_penalty (float, optional): The diversity penalty for diverse beam search. Defaults to 0.0.
            do_sample (bool, optional): Whether to use sampling instead of beam search. Defaults to False.
            temperature (float, optional): The temperature value for sampling. Defaults to 1.0.
            top_k (int, optional): The value for top-k sampling. Defaults to 50.
            top_p (float, optional): The value for top-p (nucleus) sampling. Defaults to 1.0.

        Returns:
            GenerationOutputs: The generation outputs.
        """
        outputs = super().generate(
            pixel_values=pixel_values,
            prefix_input_ids=prefix_input_ids,
            suffix_input_ids=suffix_input_ids,
            num_beams=num_beams,
            decoder_start_token_id=decoder_start_token_id,
            decoder_end_token_id=decoder_end_token_id,
            num_return_sequences=num_return_sequences,
            min_gen_seq_length=min_gen_seq_length,
            max_gen_seq_length=max_gen_seq_length,
            repetition_penalty=repetition_penalty,
            no_repeat_ngram_size=no_repeat_ngram_size,
            early_stopping=early_stopping,
            length_penalty=length_penalty,
            num_beam_groups=num_beam_groups,
            diversity_penalty=diversity_penalty,
            do_sample=do_sample,
            temperature=temperature,
            top_k=top_k,
            top_p=top_p,
        )

        return GenerationOutputs(
            sequences=outputs.sequences,
            sequences_scores=outputs.sequences_scores,
        )



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import json
import torch
import torch.nn.functional as F
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
import diffusers.schedulers as schedulers
from transformers import CLIPTextConfig, CLIPTextModel
from diffusers.schedulers import SchedulerMixin
from diffusers.models.attention_processor import (
    AttnAddedKVProcessor,
    AttnAddedKVProcessor2_0,
    LoRAAttnAddedKVProcessor,
    LoRAAttnProcessor,
    LoRAAttnProcessor2_0,
    SlicedAttnAddedKVProcessor,
)
from diffusers.models import (
    ControlNetModel,
    UNet2DModel,
    UNet2DConditionModel,
    AutoencoderKL,
)
from diffusers.pipelines import (
    StableDiffusionPipeline,
    StableDiffusionControlNetPipeline,
)
from diffusers.pipelines.blip_diffusion.modeling_blip2 import (
    Blip2Config,
    Blip2QFormerModel,
    BaseModelOutputWithPoolingAndCrossAttentions,
)
from diffusers.pipelines.blip_diffusion.modeling_ctx_clip import ContextCLIPTextModel
from unitorch.models import (
    GenericModel,
    GenericOutputs,
    QuantizationConfig,
    QuantizationMixin,
)
from unitorch.models.diffusers.modeling_stable import compute_snr


class Blip2QFormerModelV2(Blip2QFormerModel):
    def forward(
        self,
        input_ids,
        attention_mask,
        pixel_values,
        head_mask=None,
        encoder_hidden_states=None,
        encoder_attention_mask=None,
        past_key_values=None,
        use_cache=None,
        output_attentions=None,
        output_hidden_states=None,
        return_dict=None,
    ):
        batch_size = input_ids.shape[0]
        query_atts = torch.ones(
            (batch_size, self.query_tokens.size()[1]), dtype=torch.long
        ).to(self.device)
        attention_mask = torch.cat([query_atts, attention_mask], dim=1)

        output_attentions = (
            output_attentions
            if output_attentions is not None
            else self.config.output_attentions
        )
        output_hidden_states = (
            output_hidden_states
            if output_hidden_states is not None
            else self.config.output_hidden_states
        )
        return_dict = (
            return_dict if return_dict is not None else self.config.use_return_dict
        )

        # past_key_values_length
        past_key_values_length = (
            past_key_values[0][0].shape[2] - self.config.query_length
            if past_key_values is not None
            else 0
        )

        query_length = self.query_tokens.shape[1]

        embedding_output = self.embeddings(
            input_ids=input_ids,
            query_embeds=self.query_tokens,
            past_key_values_length=past_key_values_length,
        )

        # embedding_output = self.layernorm(query_embeds)
        # embedding_output = self.dropout(embedding_output)

        input_shape = embedding_output.size()[:-1]
        batch_size, seq_length = input_shape
        device = embedding_output.device

        image_embeds_frozen = self.visual_encoder(pixel_values).last_hidden_state
        # image_embeds_frozen = torch.ones_like(image_embeds_frozen)
        encoder_hidden_states = image_embeds_frozen

        if attention_mask is None:
            attention_mask = torch.ones(
                ((batch_size, seq_length + past_key_values_length)), device=device
            )

        # We can provide a self-attention mask of dimensions [batch_size, from_seq_length, to_seq_length]
        # ourselves in which case we just need to make it broadcastable to all heads.
        extended_attention_mask = self.get_extended_attention_mask(
            attention_mask, input_shape, device
        )

        # If a 2D or 3D attention mask is provided for the cross-attention
        # we need to make broadcastable to [batch_size, num_heads, seq_length, seq_length]
        if encoder_hidden_states is not None:
            if isinstance(encoder_hidden_states, list):
                encoder_batch_size, encoder_sequence_length, _ = encoder_hidden_states[
                    0
                ].size()
            else:
                (
                    encoder_batch_size,
                    encoder_sequence_length,
                    _,
                ) = encoder_hidden_states.size()
            encoder_hidden_shape = (encoder_batch_size, encoder_sequence_length)

            if isinstance(encoder_attention_mask, list):
                encoder_extended_attention_mask = [
                    self.invert_attention_mask(mask) for mask in encoder_attention_mask
                ]
            elif encoder_attention_mask is None:
                encoder_attention_mask = torch.ones(encoder_hidden_shape, device=device)
                encoder_extended_attention_mask = self.invert_attention_mask(
                    encoder_attention_mask
                )
            else:
                encoder_extended_attention_mask = self.invert_attention_mask(
                    encoder_attention_mask
                )
        else:
            encoder_extended_attention_mask = None

        # Prepare head mask if needed
        # 1.0 in head_mask indicate we keep the head
        # attention_probs has shape bsz x n_heads x N x N
        # input head_mask has shape [num_heads] or [num_hidden_layers x num_heads]
        # and head_mask is converted to shape [num_hidden_layers x batch x num_heads x seq_length x seq_length]
        head_mask = self.get_head_mask(
            head_mask, self.config.qformer_config.num_hidden_layers
        )

        encoder_outputs = self.encoder(
            embedding_output,
            attention_mask=extended_attention_mask,
            head_mask=head_mask,
            encoder_hidden_states=encoder_hidden_states,
            encoder_attention_mask=encoder_extended_attention_mask,
            past_key_values=past_key_values,
            use_cache=use_cache,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
            query_length=query_length,
        )
        sequence_output = encoder_outputs[0]
        pooled_output = sequence_output[:, 0, :]

        if not return_dict:
            return self.proj_layer(sequence_output[:, :query_length, :])

        return BaseModelOutputWithPoolingAndCrossAttentions(
            last_hidden_state=sequence_output,
            pooler_output=pooled_output,
            past_key_values=encoder_outputs.past_key_values,
            hidden_states=encoder_outputs.hidden_states,
            attentions=encoder_outputs.attentions,
            cross_attentions=encoder_outputs.cross_attentions,
        )


class Blip2ForText2ImageGeneration(GenericModel, QuantizationMixin):
    """
    Blip2ForText2ImageGeneration is a class that represents a model for text-to-image generation using the Blip2 architecture.

    Args:
        config_path (str): The path to the configuration file for the Blip2 model.
        text_config_path (str): The path to the configuration file for the text model.
        qformer_config_path (str): The path to the configuration file for the qformer model.
        vae_config_path (str): The path to the configuration file for the VAE model.
        scheduler_config_path (str): The path to the configuration file for the scheduler.
        quant_config_path (Optional[str]): The path to the configuration file for quantization (default: None).
        image_size (Optional[int]): The size of the input image (default: None).
        in_channels (Optional[int]): The number of input channels (default: None).
        out_channels (Optional[int]): The number of output channels (default: None).
        num_train_timesteps (Optional[int]): The number of training timesteps (default: 1000).
        num_infer_timesteps (Optional[int]): The number of inference timesteps (default: 50).
        freeze_vae_encoder (Optional[bool]): Whether to freeze the VAE encoder (default: True).
        freeze_text_encoder (Optional[bool]): Whether to freeze the text encoder (default: True).
        snr_gamma (Optional[float]): The SNR gamma value (default: 5.0).
        prior_loss_weight (Optional[float]): The weight for the prior loss (default: 1.0).
        lora_r (Optional[int]): The value of lora_r (default: None).
        seed (Optional[int]): The random seed (default: 1123).

    Attributes:
        unet (UNet2DConditionModel): The UNet model.
        text (ContextCLIPTextModel): The text model.
        qformer (Blip2QFormerModelV2): The qformer model.
        vae (AutoencoderKL): The VAE model.
        scheduler (SchedulerMixin): The scheduler.
        vae_scale_factor (int): The scale factor for the VAE.
        quant_config (QuantizationConfig): The quantization configuration.
        pipeline (StableDiffusionPipeline): The diffusion pipeline.

    Methods:
        enable_lora: Enables LoRA attention.
        forward: Performs forward pass (not implemented).
        generate: Generates images based on input prompts.

    """

    prefix_keys_in_state_dict = {
        # unet weights
        "^conv_in.*": "unet.",
        "^conv_norm_out.*": "unet.",
        "^conv_out.*": "unet.",
        "^time_embedding.*": "unet.",
        "^up_blocks.*": "unet.",
        "^mid_block.*": "unet.",
        "^down_blocks.*": "unet.",
        # text weights
        "^text_model.*": "text.",
        # vae weights
        "^encoder.*": "vae.",
        "^decoder.*": "vae.",
        "^post_quant_conv.*": "vae.",
        "^quant_conv.*": "vae.",
    }

    # replace_keys_in_state_dict = {
    #     "\.query\.": ".to_q.",
    #     "\.key\.": ".to_k.",
    #     "\.value\.": ".to_v.",
    #     "\.proj_attn\.": ".to_out.0.",
    # }

    def __init__(
        self,
        config_path: str,
        text_config_path: str,
        qformer_config_path: str,
        vae_config_path: str,
        scheduler_config_path: str,
        quant_config_path: Optional[str] = None,
        image_size: Optional[int] = None,
        in_channels: Optional[int] = None,
        out_channels: Optional[int] = None,
        num_train_timesteps: Optional[int] = 1000,
        num_infer_timesteps: Optional[int] = 50,
        freeze_vae_encoder: Optional[bool] = True,
        freeze_text_encoder: Optional[bool] = True,
        snr_gamma: Optional[float] = 5.0,
        prior_loss_weight: Optional[float] = 1.0,
        lora_r: Optional[int] = None,
        seed: Optional[int] = 1123,
    ):
        super().__init__()
        self.seed = seed
        self.num_train_timesteps = num_train_timesteps
        self.num_infer_timesteps = num_infer_timesteps
        self.image_size = image_size
        self.snr_gamma = snr_gamma
        self.prior_loss_weight = prior_loss_weight

        config_dict = json.load(open(config_path))
        if image_size is not None:
            config_dict.update({"sample_size": image_size})
        if in_channels is not None:
            config_dict.update({"in_channels": in_channels})
        if out_channels is not None:
            config_dict.update({"out_channels": out_channels})
        self.unet = UNet2DConditionModel.from_config(config_dict)

        text_config = CLIPTextConfig.from_json_file(text_config_path)
        self.text = ContextCLIPTextModel(text_config)

        qformer_config = Blip2Config.from_json_file(qformer_config_path)
        self.qformer = Blip2QFormerModelV2(qformer_config)

        vae_config_dict = json.load(open(vae_config_path))
        self.vae = AutoencoderKL.from_config(vae_config_dict)

        scheduler_config_dict = json.load(open(scheduler_config_path))
        scheduler_class_name = scheduler_config_dict.get("_class_name", "DDPMScheduler")
        assert hasattr(schedulers, scheduler_class_name)
        scheduler_class = getattr(schedulers, scheduler_class_name)
        assert issubclass(scheduler_class, SchedulerMixin)
        scheduler_config_dict["num_train_timesteps"] = num_train_timesteps
        self.scheduler = scheduler_class.from_config(scheduler_config_dict)

        self.vae_scale_factor = 2 ** (len(self.vae.config.block_out_channels) - 1)

        if freeze_vae_encoder:
            for param in self.vae.parameters():
                param.requires_grad = False

        if freeze_text_encoder:
            for param in self.text.parameters():
                param.requires_grad = False

        if quant_config_path is not None:
            self.quant_config = QuantizationConfig.from_json_file(quant_config_path)
            self.quantize(self.quant_config, ignore_modules=["lm_head", "unet", "vae"])

        if lora_r is not None:
            for param in self.unet.parameters():
                param.requires_grad = False
            self.enable_lora(lora_r=lora_r)

        self.scheduler.set_timesteps(num_inference_steps=self.num_infer_timesteps)
        self.pipeline = StableDiffusionPipeline(
            vae=self.vae,
            text_encoder=self.text,
            unet=self.unet,
            scheduler=self.scheduler,
            tokenizer=None,
            safety_checker=None,
            feature_extractor=None,
        )
        self.pipeline.set_progress_bar_config(disable=True)

    def enable_lora(self, lora_r: Optional[int] = 4):
        """
        Enables LoRA attention in the UNet model.

        Args:
            lora_r (Optional[int]): The value of lora_r (default: 4).

        """
        lora_attn_procs = {}
        for name, attn_processor in self.unet.attn_processors.items():
            cross_attention_dim = (
                None
                if name.endswith("attn1.processor")
                else self.unet.config.cross_attention_dim
            )
            if name.startswith("mid_block"):
                hidden_size = self.unet.config.block_out_channels[-1]
            elif name.startswith("up_blocks"):
                block_id = int(name[len("up_blocks.")])
                hidden_size = list(reversed(self.unet.config.block_out_channels))[
                    block_id
                ]
            elif name.startswith("down_blocks"):
                block_id = int(name[len("down_blocks.")])
                hidden_size = self.unet.config.block_out_channels[block_id]

            if isinstance(
                attn_processor,
                (
                    AttnAddedKVProcessor,
                    SlicedAttnAddedKVProcessor,
                    AttnAddedKVProcessor2_0,
                ),
            ):
                lora_attn_processor_class = LoRAAttnAddedKVProcessor
            else:
                lora_attn_processor_class = (
                    LoRAAttnProcessor2_0
                    if hasattr(F, "scaled_dot_product_attention")
                    else LoRAAttnProcessor
                )

            module = lora_attn_processor_class(
                hidden_size=hidden_size,
                cross_attention_dim=cross_attention_dim,
                rank=lora_r,
            )

            lora_attn_procs[name] = module

        self.unet.set_attn_processor(lora_attn_procs)

    def forward(
        self,
    ):
        """
        Performs a forward pass through the model.

        Raises:
            NotImplementedError: This method is not implemented.

        """
        raise NotImplementedError

    def generate(
        self,
        input_ids: torch.Tensor,
        negative_input_ids: torch.Tensor,
        refer_input_ids: torch.Tensor,
        refer_pixel_values: torch.Tensor,
        attention_mask: Optional[torch.Tensor] = None,
        negative_attention_mask: Optional[torch.Tensor] = None,
        refer_attention_mask: Optional[torch.Tensor] = None,
        height: Optional[int] = 512,
        width: Optional[int] = 512,
        ctx_begin_pos: Optional[int] = 2,
        guidance_scale: Optional[float] = 7.5,
    ):
        """
        Generates images based on input prompts.

        Args:
            input_ids (torch.Tensor): The input IDs.
            negative_input_ids (torch.Tensor): The negative input IDs.
            refer_input_ids (torch.Tensor): The reference input IDs.
            refer_pixel_values (torch.Tensor): The reference pixel values.
            attention_mask (Optional[torch.Tensor]): The attention mask (default: None).
            negative_attention_mask (Optional[torch.Tensor]): The negative attention mask (default: None).
            refer_attention_mask (Optional[torch.Tensor]): The reference attention mask (default: None).
            height (Optional[int]): The height of the generated images (default: 512).
            width (Optional[int]): The width of the generated images (default: 512).
            ctx_begin_pos (Optional[int]): The context begin position (default: 2).
            guidance_scale (Optional[float]): The guidance scale (default: 7.5).

        Returns:
            GenericOutputs: The generated images.

        """
        batch_size = input_ids.shape[0]
        qformer_embeds = self.qformer(
            refer_input_ids,
            refer_attention_mask,
            refer_pixel_values,
            return_dict=False,
        )
        prompt_embeds = self.text(
            input_ids=input_ids,
            # attention_mask,
            ctx_embeddings=qformer_embeds,
            ctx_begin_pos=[ctx_begin_pos] * batch_size,
        )[0]
        negative_prompt_embeds = self.text(
            input_ids=negative_input_ids,
            # negative_attention_mask,
        )[0]

        images = self.pipeline(
            prompt_embeds=prompt_embeds,
            negative_prompt_embeds=negative_prompt_embeds,
            generator=torch.Generator(device=self.pipeline.device).manual_seed(
                self.seed
            ),
            height=height,
            width=width,
            guidance_scale=guidance_scale,
            output_type="np.array",
        ).images

        return GenericOutputs(images=torch.from_numpy(images))


class Blip2ControlNetForText2ImageGeneration(GenericModel, QuantizationMixin):
    """
    Blip2ControlNetForText2ImageGeneration is a class that represents a control network for text-to-image generation using the Blip2 model.

    Args:
        config_path (str): The path to the configuration file for the Blip2 model.
        text_config_path (str): The path to the configuration file for the text model.
        qformer_config_path (str): The path to the configuration file for the qformer model.
        vae_config_path (str): The path to the configuration file for the VAE model.
        controlnet_config_path (str): The path to the configuration file for the control network model.
        scheduler_config_path (str): The path to the configuration file for the scheduler model.
        quant_config_path (Optional[str]): The path to the configuration file for quantization (default: None).
        image_size (Optional[int]): The size of the input image (default: None).
        in_channels (Optional[int]): The number of input channels (default: None).
        out_channels (Optional[int]): The number of output channels (default: None).
        num_train_timesteps (Optional[int]): The number of training timesteps (default: 1000).
        num_infer_timesteps (Optional[int]): The number of inference timesteps (default: 50).
        freeze_vae_encoder (Optional[bool]): Whether to freeze the VAE encoder (default: True).
        freeze_text_encoder (Optional[bool]): Whether to freeze the text encoder (default: True).
        snr_gamma (Optional[float]): The SNR gamma value (default: 5.0).
        prior_loss_weight (Optional[float]): The weight for the prior loss (default: 1.0).
        lora_r (Optional[int]): The LORA rank (default: None).
        seed (Optional[int]): The random seed (default: 1123).

    Attributes:
        unet (UNet2DConditionModel): The UNet model.
        text (ContextCLIPTextModel): The text model.
        qformer (Blip2QFormerModelV2): The qformer model.
        vae (AutoencoderKL): The VAE model.
        controlnet (ControlNetModel): The control network model.
        scheduler (SchedulerMixin): The scheduler model.
        vae_scale_factor (int): The scale factor for the VAE.

    Methods:
        enable_lora(lora_r: Optional[int] = 4): Enables LoRA attention for the UNet model.
        forward(): Performs the forward pass of the model.
        generate(input_ids: torch.Tensor, negative_input_ids: torch.Tensor, refer_input_ids: torch.Tensor, refer_pixel_values: torch.Tensor, condition_pixel_values: torch.Tensor, attention_mask: Optional[torch.Tensor] = None, negative_attention_mask: Optional[torch.Tensor] = None, refer_attention_mask: Optional[torch.Tensor] = None, height: Optional[int] = 512, width: Optional[int] = 512, ctx_begin_pos: Optional[int] = 2, guidance_scale: Optional[float] = 7.5): Generates images based on the input prompts.

    """

    prefix_keys_in_state_dict = {
        # unet weights
        "^conv_in.*": "unet.",
        "^conv_norm_out.*": "unet.",
        "^conv_out.*": "unet.",
        "^time_embedding.*": "unet.",
        "^up_blocks.*": "unet.",
        "^mid_block.*": "unet.",
        "^down_blocks.*": "unet.",
        # text weights
        "^text_model.*": "text.",
        # vae weights
        "^encoder.*": "vae.",
        "^decoder.*": "vae.",
        "^post_quant_conv.*": "vae.",
        "^quant_conv.*": "vae.",
    }

    # replace_keys_in_state_dict = {
    #     "\.query\.": ".to_q.",
    #     "\.key\.": ".to_k.",
    #     "\.value\.": ".to_v.",
    #     "\.proj_attn\.": ".to_out.0.",
    # }

    def __init__(
        self,
        config_path: str,
        text_config_path: str,
        qformer_config_path: str,
        vae_config_path: str,
        controlnet_config_path: str,
        scheduler_config_path: str,
        quant_config_path: Optional[str] = None,
        image_size: Optional[int] = None,
        in_channels: Optional[int] = None,
        out_channels: Optional[int] = None,
        num_train_timesteps: Optional[int] = 1000,
        num_infer_timesteps: Optional[int] = 50,
        freeze_vae_encoder: Optional[bool] = True,
        freeze_text_encoder: Optional[bool] = True,
        snr_gamma: Optional[float] = 5.0,
        prior_loss_weight: Optional[float] = 1.0,
        lora_r: Optional[int] = None,
        seed: Optional[int] = 1123,
    ):
        """
        Initializes the Blip2 model.

        Args:
            config_path (str): The path to the configuration file.
            text_config_path (str): The path to the text configuration file.
            qformer_config_path (str): The path to the qformer configuration file.
            vae_config_path (str): The path to the VAE configuration file.
            controlnet_config_path (str): The path to the controlnet configuration file.
            scheduler_config_path (str): The path to the scheduler configuration file.
            quant_config_path (str, optional): The path to the quantization configuration file. Defaults to None.
            image_size (int, optional): The size of the image. Defaults to None.
            in_channels (int, optional): The number of input channels. Defaults to None.
            out_channels (int, optional): The number of output channels. Defaults to None.
            num_train_timesteps (int, optional): The number of training timesteps. Defaults to 1000.
            num_infer_timesteps (int, optional): The number of inference timesteps. Defaults to 50.
            freeze_vae_encoder (bool, optional): Whether to freeze the VAE encoder. Defaults to True.
            freeze_text_encoder (bool, optional): Whether to freeze the text encoder. Defaults to True.
            snr_gamma (float, optional): The SNR gamma value. Defaults to 5.0.
            prior_loss_weight (float, optional): The weight of the prior loss. Defaults to 1.0.
            lora_r (int, optional): The LORA R value. Defaults to None.
            seed (int, optional): The seed value. Defaults to 1123.
        """
        super().__init__()
        self.seed = seed
        self.num_train_timesteps = num_train_timesteps
        self.num_infer_timesteps = num_infer_timesteps
        self.image_size = image_size
        self.snr_gamma = snr_gamma
        self.prior_loss_weight = prior_loss_weight

        # Load configuration files
        config_dict = json.load(open(config_path))
        if image_size is not None:
            config_dict.update({"sample_size": image_size})
        if in_channels is not None:
            config_dict.update({"in_channels": in_channels})
        if out_channels is not None:
            config_dict.update({"out_channels": out_channels})
        self.unet = UNet2DConditionModel.from_config(config_dict)

        text_config = CLIPTextConfig.from_json_file(text_config_path)
        self.text = ContextCLIPTextModel(text_config)

        qformer_config = Blip2Config.from_json_file(qformer_config_path)
        self.qformer = Blip2QFormerModelV2(qformer_config)

        vae_config_dict = json.load(open(vae_config_path))
        self.vae = AutoencoderKL.from_config(vae_config_dict)

        controlnet_config_dict = json.load(open(controlnet_config_path))
        self.controlnet = ControlNetModel.from_config(controlnet_config_dict)

        scheduler_config_dict = json.load(open(scheduler_config_path))
        scheduler_class_name = scheduler_config_dict.get("_class_name", "DDPMScheduler")
        assert hasattr(schedulers, scheduler_class_name)
        scheduler_class = getattr(schedulers, scheduler_class_name)
        assert issubclass(scheduler_class, SchedulerMixin)
        scheduler_config_dict["num_train_timesteps"] = num_train_timesteps
        self.scheduler = scheduler_class.from_config(scheduler_config_dict)

        self.vae_scale_factor = 2 ** (len(self.vae.config.block_out_channels) - 1)

        # Freeze VAE encoder if specified
        if freeze_vae_encoder:
            for param in self.vae.parameters():
                param.requires_grad = False

        # Freeze text encoder if specified
        if freeze_text_encoder:
            for param in self.text.parameters():
                param.requires_grad = False

        # Load quantization configuration if specified
        if quant_config_path is not None:
            self.quant_config = QuantizationConfig.from_json_file(quant_config_path)
            self.quantize(self.quant_config, ignore_modules=["lm_head", "unet", "vae"])

        # Enable LORA if specified
        if lora_r is not None:
            for param in self.unet.parameters():
                param.requires_grad = False
            self.enable_lora(lora_r=lora_r)

        self.scheduler.set_timesteps(num_inference_steps=self.num_infer_timesteps)
        self.pipeline = StableDiffusionControlNetPipeline(
            vae=self.vae,
            text_encoder=self.text,
            unet=self.unet,
            controlnet=self.controlnet,
            scheduler=self.scheduler,
            tokenizer=None,
            safety_checker=None,
            feature_extractor=None,
        )
        self.pipeline.set_progress_bar_config(disable=True)

    def enable_lora(self, lora_r: Optional[int] = 4):
        """
        Enable LoRA (Local Rank Attention) for the UNet model.

        Args:
            lora_r (Optional[int]): The rank parameter for LoRA. Defaults to 4.

        Returns:
            None
        """
        lora_attn_procs = {}
        for name, attn_processor in self.unet.attn_processors.items():
            cross_attention_dim = (
                None
                if name.endswith("attn1.processor")
                else self.unet.config.cross_attention_dim
            )
            if name.startswith("mid_block"):
                hidden_size = self.unet.config.block_out_channels[-1]
            elif name.startswith("up_blocks"):
                block_id = int(name[len("up_blocks.")])
                hidden_size = list(reversed(self.unet.config.block_out_channels))[
                    block_id
                ]
            elif name.startswith("down_blocks"):
                block_id = int(name[len("down_blocks.")])
                hidden_size = self.unet.config.block_out_channels[block_id]

            if isinstance(
                attn_processor,
                (
                    AttnAddedKVProcessor,
                    SlicedAttnAddedKVProcessor,
                    AttnAddedKVProcessor2_0,
                ),
            ):
                lora_attn_processor_class = LoRAAttnAddedKVProcessor
            else:
                lora_attn_processor_class = (
                    LoRAAttnProcessor2_0
                    if hasattr(F, "scaled_dot_product_attention")
                    else LoRAAttnProcessor
                )

            module = lora_attn_processor_class(
                hidden_size=hidden_size,
                cross_attention_dim=cross_attention_dim,
                rank=lora_r,
            )

            lora_attn_procs[name] = module

        self.unet.set_attn_processor(lora_attn_procs)

    def forward(
        self,
    ):
        """
        Performs the forward pass of the model.

        Raises:
            NotImplementedError: This method should be implemented in a subclass.
        """
        raise NotImplementedError

    def generate(
        self,
        input_ids: torch.Tensor,
        negative_input_ids: torch.Tensor,
        refer_input_ids: torch.Tensor,
        refer_pixel_values: torch.Tensor,
        condition_pixel_values: torch.Tensor,
        attention_mask: Optional[torch.Tensor] = None,
        negative_attention_mask: Optional[torch.Tensor] = None,
        refer_attention_mask: Optional[torch.Tensor] = None,
        height: Optional[int] = 512,
        width: Optional[int] = 512,
        ctx_begin_pos: Optional[int] = 2,
        guidance_scale: Optional[float] = 7.5,
        controlnet_conditioning_scale: Optional[float] = 1.0,
    ):
        """
        Generates images based on the given input_ids, negative_input_ids, refer_input_ids,
        refer_pixel_values, condition_pixel_values, and other optional parameters.

        Args:
            input_ids (torch.Tensor): The input tensor representing the prompt.
            negative_input_ids (torch.Tensor): The input tensor representing the negative prompt.
            refer_input_ids (torch.Tensor): The input tensor representing the reference prompt.
            refer_pixel_values (torch.Tensor): The tensor representing the pixel values of the reference image.
            condition_pixel_values (torch.Tensor): The tensor representing the pixel values of the condition image.
            attention_mask (Optional[torch.Tensor], optional): The attention mask tensor. Defaults to None.
            negative_attention_mask (Optional[torch.Tensor], optional): The attention mask tensor for negative_input_ids. Defaults to None.
            refer_attention_mask (Optional[torch.Tensor], optional): The attention mask tensor for refer_input_ids. Defaults to None.
            height (Optional[int], optional): The height of the generated images. Defaults to 512.
            width (Optional[int], optional): The width of the generated images. Defaults to 512.
            ctx_begin_pos (Optional[int], optional): The beginning position of the context embeddings. Defaults to 2.
            guidance_scale (Optional[float], optional): The scale factor for guidance. Defaults to 7.5.

        Returns:
            GenericOutputs: The generated images.
        """
        batch_size = input_ids.shape[0]
        qformer_embeds = self.qformer(
            refer_input_ids,
            refer_attention_mask,
            refer_pixel_values,
            return_dict=False,
        )
        prompt_embeds = self.text(
            input_ids=input_ids,
            # attention_mask,
            ctx_embeddings=qformer_embeds,
            ctx_begin_pos=[ctx_begin_pos] * batch_size,
        )[0]
        negative_prompt_embeds = self.text(
            input_ids=negative_input_ids,
            # negative_attention_mask,
        )[0]

        images = self.pipeline(
            image=condition_pixel_values,
            prompt_embeds=prompt_embeds,
            negative_prompt_embeds=negative_prompt_embeds,
            generator=torch.Generator(device=self.pipeline.device).manual_seed(
                self.seed
            ),
            height=height,
            width=width,
            guidance_scale=guidance_scale,
            controlnet_conditioning_scale=float(controlnet_conditioning_scale),
            output_type="np.array",
        ).images

        return GenericOutputs(images=torch.from_numpy(images))



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import os
import json
import logging
import torch
import torch.nn as nn
import torch.nn.functional as F
import transformers
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from transformers import BloomModel, BloomConfig, BloomForCausalLM
from unitorch.utils.decorators import replace
from unitorch.models import GenericModel, GenericOutputs


class BloomForClassification(GenericModel):
    """
    A classification model based on the Bloom architecture.
    """

    prefix_keys_in_state_dict = {"^(?!transformer\.).*": "transformer."}

    def __init__(
        self,
        config_path: str,
        num_classes: Optional[int] = 1,
        hidden_dropout_prob: Optional[float] = 0.1,
        gradient_checkpointing: Optional[bool] = False,
    ):
        """
        Initializes a new instance of the BloomForClassification model.

        Args:
            config_path (str): The path to the configuration file for the Bloom model.
            num_classes (Optional[int]): The number of output classes for classification. Defaults to 1.
            hidden_dropout_prob (Optional[float]): The dropout probability for the hidden layers. Defaults to 0.1.
            gradient_checkpointing (Optional[bool]): Whether to use gradient checkpointing. Defaults to False.
        """
        super().__init__()
        self.config = BloomConfig.from_json_file(config_path)
        self.config.gradient_checkpointing = gradient_checkpointing
        self.transformer = BloomModel(self.config)
        self.dropout = nn.Dropout(hidden_dropout_prob)
        self.classifier = nn.Linear(self.config.hidden_size, num_classes)
        self.init_weights()

    def forward(
        self,
        input_ids: torch.Tensor,
        attention_mask: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        """
        Forward pass of the BloomForClassification model.

        Args:
            input_ids (torch.Tensor): The input token IDs.
            attention_mask (Optional[torch.Tensor]): The attention mask tensor. Defaults to None.

        Returns:
            (torch.Tensor):The output logits for classification.
        """
        outputs = self.transformer(
            input_ids,
            attention_mask=attention_mask,
        )[0]
        pooled_output = outputs[:, -1]
        pooled_output = self.dropout(pooled_output)
        logits = self.classifier(pooled_output)
        return logits


class BloomForPretrain(GenericModel):
    """
    A pretraining model based on the Bloom architecture.
    """

    prefix_keys_in_state_dict = {"^(?!transformer\.).*": "transformer."}

    def __init__(
        self,
        config_path: str,
        gradient_checkpointing: Optional[bool] = False,
    ):
        """
        Initializes a new instance of the BloomForPretrain model.

        Args:
            config_path (str): The path to the configuration file for the Bloom model.
            gradient_checkpointing (Optional[bool]): Whether to use gradient checkpointing. Defaults to False.
        """
        super().__init__()
        self.config = BloomConfig.from_json_file(config_path)
        self.config.gradient_checkpointing = gradient_checkpointing
        self.transformer = BloomModel(self.config)
        self.lm_head = nn.Linear(
            self.config.hidden_size, self.config.vocab_size, bias=False
        )
        self.init_weights()

    def forward(
        self,
        input_ids: torch.Tensor,
        attention_mask: torch.Tensor,
        input_ids_label: torch.Tensor,
        attention_mask_label: torch.Tensor,
    ) -> torch.Tensor:
        """
        Forward pass of the BloomForPretrain model.

        Args:
            input_ids (torch.Tensor): The input token IDs.
            attention_mask (torch.Tensor): The attention mask tensor.
            input_ids_label (torch.Tensor): The input token IDs for the labeled data.
            attention_mask_label (torch.Tensor): The attention mask tensor for the labeled data.

        Returns:
            (torch.Tensor):The computed loss value.
        """
        outputs = self.transformer(
            input_ids=input_ids,
            attention_mask=attention_mask,
        )
        predict_logits = self.lm_head(outputs[0])
        batch_size, seq_len, num_classes = predict_logits.size()
        logits = predict_logits.contiguous().view(batch_size * seq_len, num_classes)
        targets = input_ids_label.contiguous().view(-1).long()
        masks = attention_mask_label.contiguous().view(-1)
        loss = nn.CrossEntropyLoss(reduction="none")(logits, targets)
        loss = loss * masks.float()
        loss = loss.contiguous().view(batch_size, seq_len).sum(1) / torch.max(
            masks.contiguous().view(batch_size, seq_len).float().sum(1),
            torch.ones(batch_size).to(masks.device),
        )
        loss = torch.mean(loss)
        return loss


class BloomForGeneration(GenericModel):
    """
    A generation model based on the Bloom architecture.
    """

    prefix_keys_in_state_dict = {"^(?!model\.).*": "model.transformer."}

    def __init__(
        self,
        config_path: str,
        gradient_checkpointing: Optional[bool] = False,
    ):
        """
        Initializes a new instance of the BloomForGeneration model.

        Args:
            config_path (str): The path to the configuration file for the Bloom model.
            gradient_checkpointing (Optional[bool]): Whether to use gradient checkpointing. Defaults to False.
        """
        super().__init__()
        self.config = BloomConfig.from_json_file(config_path)
        self.config.gradient_checkpointing = gradient_checkpointing
        self.model = BloomForCausalLM(self.config)
        self.init_weights()

    def forward(
        self,
        input_ids: torch.Tensor,
        attention_mask: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        """
        Forward pass of the BloomForGeneration model.

        Args:
            input_ids (torch.Tensor): The input token IDs.
            attention_mask (Optional[torch.Tensor]): The attention mask tensor. Defaults to None.

        Returns:
            (torch.Tensor):The output logits.
        """
        outputs = self.model(
            input_ids,
            attention_mask=attention_mask,
            return_dict=True,
        )
        logits = outputs.logits
        return logits

    @torch.no_grad()
    def generate(
        self,
        input_ids: torch.Tensor,
        num_beams: Optional[int] = 5,
        decoder_start_token_id: Optional[int] = 1,
        decoder_end_token_id: Optional[Union[int, List[int]]] = 2,
        num_return_sequences: Optional[int] = 1,
        min_gen_seq_length: Optional[int] = 0,
        max_gen_seq_length: Optional[int] = 48,
        repetition_penalty: Optional[float] = 1.0,
        no_repeat_ngram_size: Optional[int] = 0,
        early_stopping: Optional[bool] = True,
        length_penalty: Optional[float] = 1.0,
        num_beam_groups: Optional[int] = 1,
        diversity_penalty: Optional[float] = 0.0,
        do_sample: Optional[bool] = False,
        temperature: Optional[float] = 1.0,
        top_k: Optional[int] = 50,
        top_p: Optional[float] = 1.0,
    ) -> GenericOutputs:
        """
        Generate sequences using the BloomForGeneration model.

        Args:
            input_ids (torch.Tensor): The input token IDs.
            num_beams (Optional[int]): The number of beams for beam search. Defaults to 5.
            decoder_start_token_id (Optional[int]): The ID of the start token for decoding. Defaults to 1.
            decoder_end_token_id (Optional[int]): The ID of the end token for decoding. Defaults to 2.
            num_return_sequences (Optional[int]): The number of generated sequences to return. Defaults to 1.
            min_gen_seq_length (Optional[int]): The minimum length of the generated sequences. Defaults to 0.
            max_gen_seq_length (Optional[int]): The maximum length of the generated sequences. Defaults to 48.
            repetition_penalty (Optional[float]): The penalty for repeated n-grams. Defaults to 1.0.
            no_repeat_ngram_size (Optional[int]): The size of n-grams to prevent repetition. Defaults to 0.
            early_stopping (Optional[bool]): Whether to stop generation early based on specified conditions. Defaults to True.
            length_penalty (Optional[float]): The penalty for longer sequences. Defaults to 1.0.
            num_beam_groups (Optional[int]): The number of beam groups for diverse beam search. Defaults to 1.
            diversity_penalty (Optional[float]): The penalty for diverse beam search. Defaults to 0.0.
            do_sample (Optional[bool]): Whether to use sampling for generation. Defaults to False.
            temperature (Optional[float]): The temperature for sampling. Defaults to 1.0.
            top_k (Optional[int]): The number of top-k tokens to consider for sampling. Defaults to 50.
            top_p (Optional[float]): The cumulative probability for top-p sampling. Defaults to 1.0.

        Returns:
            GenericOutputs: The generated sequences and their scores.
        """
        input_seq_length = input_ids.size(1)
        outputs = self.model.generate(
            input_ids,
            max_length=max_gen_seq_length + input_seq_length,
            min_length=min_gen_seq_length + input_seq_length,
            num_beams=num_beams,
            do_sample=do_sample,
            decoder_start_token_id=decoder_start_token_id,
            no_repeat_ngram_size=no_repeat_ngram_size,
            early_stopping=early_stopping,
            length_penalty=length_penalty,
            repetition_penalty=repetition_penalty,
            num_return_sequences=num_return_sequences,
            bos_token_id=decoder_start_token_id,
            eos_token_id=decoder_end_token_id,
            num_beam_groups=num_beam_groups,
            diversity_penalty=diversity_penalty,
            temperature=temperature,
            top_k=top_k,
            top_p=top_p,
            return_dict_in_generate=True,
            output_scores=True,
        )
        sequences = outputs.sequences.reshape(
            -1, num_return_sequences, outputs.sequences.size(-1)
        )
        outputs.sequences = torch.zeros(
            sequences.size(0), num_return_sequences, max_gen_seq_length
        ).to(device=sequences.device)
        outputs.sequences[:, :, : sequences.size(-1) - input_seq_length].copy_(
            sequences[:, :, input_seq_length : sequences.size(-1)]
        )

        if num_return_sequences == 1:
            outputs.sequences = outputs.sequences.reshape(-1, max_gen_seq_length)

        return GenericOutputs(
            sequences=outputs.sequences.long(),
            sequences_scores=outputs.sequences_scores,
        )



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import torch
import torch.nn as nn
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from unitorch.scores import (
    accuracy_score,
    recall_score,
    f1_score,
    bleu_score,
    map_score,
    map50_score,
    rouge1_score,
    rouge2_score,
    rougel_score,
    roc_auc_score,
    ndcg_score,
    matthews_corrcoef,
    pearsonr,
    auc,
    precision_recall_curve,
)
from unitorch.cli import add_default_section_for_init, register_score
from unitorch.cli.models import (
    ModelOutputs,
    ModelTargets,
    ClassificationOutputs,
    ClassificationTargets,
    RankingOutputs,
    RankingTargets,
    GenerationOutputs,
    GenerationTargets,
    DetectionOutputs,
    DetectionTargets,
    SegmentationOutputs,
    SegmentationTargets,
    LossOutputs,
)


class Score(nn.Module):
    pass


@register_score("core/score/acc")
class AccuracyScore(Score):
    def __init__(self, gate: Optional[float] = 0.5):
        super().__init__()
        self.gate = gate

    @classmethod
    @add_default_section_for_init("core/score/acc")
    def from_core_configure(cls, config, **kwargs):
        pass

    def forward(
        self,
        outputs: Union[ClassificationOutputs, GenerationOutputs, SegmentationOutputs],
        targets: Union[ClassificationTargets, GenerationTargets, SegmentationTargets],
    ):
        if isinstance(outputs, GenerationOutputs):
            outputs = outputs.sequences
            outputs = outputs.view(-1, outputs.size(-1))

        if isinstance(targets, GenerationTargets):
            targets = targets.refs
            targets = targets.view(-1)

        if isinstance(outputs, SegmentationOutputs):
            outputs = outputs.outputs
            outputs = torch.cat([t.view(-1, t.size(-1)) for t in outputs])

        if isinstance(targets, SegmentationTargets):
            targets = targets.targets
            targets = torch.cat([t.view(-1) for t in targets])

        if isinstance(outputs, ClassificationOutputs):
            outputs = outputs.outputs

        if isinstance(targets, ClassificationTargets):
            targets = targets.targets

        if outputs.dim() == 2:
            outputs = (
                outputs.argmax(dim=-1)
                if outputs.size(-1) > 1
                else outputs[:, 0] > self.gate
            )

        if targets.dim() == 2 and targets.size(-1) == 1:
            targets = targets[:, 0]

        assert outputs.dim() == 1 and targets.dim() == 1

        return accuracy_score(targets, outputs)


@register_score("core/score/rec")
class RecallScore(Score):
    def __init__(self, gate: Optional[float] = 0.5):
        super().__init__()
        self.gate = gate

    @classmethod
    @add_default_section_for_init("core/score/rec")
    def from_core_configure(cls, config, **kwargs):
        pass

    def forward(
        self,
        outputs: ClassificationOutputs,
        targets: ClassificationTargets,
    ):
        if isinstance(outputs, ClassificationOutputs):
            outputs = outputs.outputs
        if isinstance(targets, ClassificationTargets):
            targets = targets.targets

        if outputs.dim() == 2:
            outputs = (
                outputs.argmax(dim=-1)
                if outputs.size(-1) > 1
                else outputs[:, 0] > self.gate
            )

        if targets.dim() == 2 and targets.size(-1) == 1:
            targets = targets[:, 0]

        assert outputs.dim() == 1 and targets.dim() == 1

        return recall_score(targets, outputs, average="micro")


@register_score("core/score/f1")
class F1Score(Score):
    def __init__(self, gate: Optional[float] = 0.5):
        super().__init__()
        self.gate = gate

    @classmethod
    @add_default_section_for_init("core/score/f1")
    def from_core_configure(cls, config, **kwargs):
        pass

    def forward(
        self,
        outputs: ClassificationOutputs,
        targets: ClassificationTargets,
    ):
        if isinstance(outputs, ClassificationOutputs):
            outputs = outputs.outputs
        if isinstance(targets, ClassificationTargets):
            targets = targets.targets

        if outputs.dim() == 2:
            outputs = (
                outputs.argmax(dim=-1)
                if outputs.size(-1) > 1
                else outputs[:, 0] > self.gate
            )

        if targets.dim() == 2 and targets.size(-1) == 1:
            targets = targets[:, 0]

        assert outputs.dim() == 1 and targets.dim() == 1

        return f1_score(targets, outputs, average="micro")


@register_score("core/score/auc")
class AUCScore(Score):
    def __init__(
        self,
    ):
        super().__init__()

    @classmethod
    @add_default_section_for_init("core/score/auc")
    def from_core_configure(cls, config, **kwargs):
        pass

    def forward(
        self,
        outputs: ClassificationOutputs,
        targets: ClassificationTargets,
    ):
        if isinstance(outputs, ClassificationOutputs):
            outputs = outputs.outputs
        if isinstance(targets, ClassificationTargets):
            targets = targets.targets

        if outputs.dim() == 2:
            outputs = outputs[:, 1] if outputs.size(-1) > 1 else outputs[:, 0]

        if targets.dim() == 2 and targets.size(-1) == 1:
            targets = targets[:, 0]

        assert outputs.dim() == 1 and targets.dim() == 1

        return roc_auc_score(targets, outputs)


@register_score("core/score/pr_auc")
class PRAUCScore(Score):
    def __init__(
        self,
    ):
        super().__init__()

    @classmethod
    @add_default_section_for_init("core/score/pr_auc")
    def from_core_configure(cls, config, **kwargs):
        pass

    def forward(
        self,
        outputs: ClassificationOutputs,
        targets: ClassificationTargets,
    ):
        if isinstance(outputs, ClassificationOutputs):
            outputs = outputs.outputs
        if isinstance(targets, ClassificationTargets):
            targets = targets.targets

        if outputs.dim() == 2:
            outputs = outputs[:, 1] if outputs.size(-1) > 1 else outputs[:, 0]

        if targets.dim() == 2 and targets.size(-1) == 1:
            targets = targets[:, 0]

        assert outputs.dim() == 1 and targets.dim() == 1
        precision, recall, _ = precision_recall_curve(targets, outputs)

        return auc(recall, precision)


@register_score("core/score/ndcg")
class NDCGScore(Score):
    def __init__(
        self,
    ):
        super().__init__()

    @classmethod
    @add_default_section_for_init("core/score/ndcg")
    def from_core_configure(cls, config, **kwargs):
        pass

    def forward(
        self,
        outputs: RankingOutputs,
        targets: RankingTargets,
    ):
        if isinstance(outputs, RankingOutputs):
            outputs = outputs.outputs
        if isinstance(targets, RankingTargets):
            masks = targets.masks
            targets = targets.targets

        outputs = outputs + (1 - masks) * (
            torch.min(outputs, -1, keepdim=True)[0] - 1e3
        )
        targets = targets * masks
        return ndcg_score(targets, outputs)


@register_score("core/score/mattcorr")
class MattCorrScore(Score):
    def __init__(
        self,
    ):
        super().__init__()

    @classmethod
    @add_default_section_for_init("core/score/mattcorr")
    def from_core_configure(cls, config, **kwargs):
        pass

    def forward(
        self,
        outputs: ClassificationOutputs,
        targets: ClassificationTargets,
    ):
        if isinstance(outputs, ClassificationOutputs):
            outputs = outputs.outputs
        if isinstance(targets, ClassificationTargets):
            targets = targets.targets

        if outputs.dim() == 2:
            outputs = (
                outputs.argmax(dim=-1)
                if outputs.size(-1) > 1
                else outputs[:, 0] > self.gate
            )

        if targets.dim() == 2 and targets.size(-1) == 1:
            targets = targets[:, 0]

        assert outputs.dim() == 1 and targets.dim() == 1

        return matthews_corrcoef(targets, outputs)


@register_score("core/score/pearsonr_corr")
class PearsonrCorrScore(Score):
    def __init__(
        self,
    ):
        super().__init__()

    @classmethod
    @add_default_section_for_init("core/score/pearsonr_corr")
    def from_core_configure(cls, config, **kwargs):
        pass

    def forward(
        self,
        outputs: ClassificationOutputs,
        targets: ClassificationTargets,
    ):
        if isinstance(outputs, ClassificationOutputs):
            outputs = outputs.outputs
        if isinstance(targets, ClassificationTargets):
            targets = targets.targets

        outputs = outputs.view(-1)
        targets = targets.view(-1)

        assert outputs.numel() == targets.numel()

        return pearsonr(targets, outputs)


@register_score("core/score/mae")
class MAEScore(Score):
    def __init__(
        self,
    ):
        super().__init__()

    @classmethod
    @add_default_section_for_init("core/score/mae")
    def from_core_configure(cls, config, **kwargs):
        pass

    def forward(
        self,
        outputs: Union[ClassificationOutputs, SegmentationOutputs],
        targets: Union[ClassificationTargets, SegmentationTargets],
    ):
        if isinstance(outputs, ClassificationOutputs):
            outputs = outputs.outputs
        if isinstance(targets, ClassificationTargets):
            targets = targets.targets
        if isinstance(outputs, SegmentationOutputs):
            outputs = torch.cat([m.view(-1) for m in outputs.masks], dim=0)
        if isinstance(targets, SegmentationTargets):
            targets = torch.cat([m.view(-1) for m in targets.targets], dim=0)

        outputs = outputs.view(-1)
        targets = targets.view(-1)

        assert outputs.numel() == targets.numel()

        score = torch.mean(torch.abs(targets - outputs))

        return -float(score)


@register_score("core/score/mse")
class MSEScore(Score):
    def __init__(
        self,
    ):
        super().__init__()

    @classmethod
    @add_default_section_for_init("core/score/mse")
    def from_core_configure(cls, config, **kwargs):
        pass

    def forward(
        self,
        outputs: Union[ClassificationOutputs, SegmentationOutputs],
        targets: Union[ClassificationTargets, SegmentationTargets],
    ):
        if isinstance(outputs, ClassificationOutputs):
            outputs = outputs.outputs
        if isinstance(targets, ClassificationTargets):
            targets = targets.targets
        if isinstance(outputs, SegmentationOutputs):
            outputs = torch.cat([m.view(-1) for m in outputs.masks], dim=0)
        if isinstance(targets, SegmentationTargets):
            targets = torch.cat([m.view(-1) for m in targets.targets], dim=0)

        outputs = outputs.view(-1)
        targets = targets.view(-1)

        assert outputs.numel() == targets.numel()

        score = torch.sqrt(torch.mean(torch.pow(targets - outputs, 2)))

        return -float(score)


@register_score("core/score/mse")
class MSEScore(Score):
    def __init__(
        self,
    ):
        super().__init__()

    @classmethod
    @add_default_section_for_init("core/score/mse")
    def from_core_configure(cls, config, **kwargs):
        pass

    def forward(
        self,
        outputs: Union[ClassificationOutputs, SegmentationOutputs],
        targets: Union[ClassificationTargets, SegmentationTargets],
    ):
        if isinstance(outputs, ClassificationOutputs):
            outputs = outputs.outputs
        if isinstance(targets, ClassificationTargets):
            targets = targets.targets
        if isinstance(outputs, SegmentationOutputs):
            outputs = torch.cat([m.view(-1) for m in outputs.masks], dim=0)
        if isinstance(targets, SegmentationTargets):
            targets = torch.cat([m.view(-1) for m in targets.targets], dim=0)

        outputs = outputs.view(-1)
        targets = targets.view(-1)

        assert outputs.numel() == targets.numel()

        return -float(torch.mean(targets - outputs))


@register_score("core/score/bleu")
class BleuScore(Score):
    def __init__(
        self,
    ):
        super().__init__()

    @classmethod
    @add_default_section_for_init("core/score/bleu")
    def from_core_configure(cls, config, **kwargs):
        pass

    def forward(
        self,
        outputs: GenerationOutputs,
        targets: GenerationTargets,
    ):
        if isinstance(outputs, GenerationOutputs):
            outputs = outputs.sequences
        if isinstance(targets, GenerationTargets):
            targets = targets.refs
        return bleu_score(
            targets.long(),
            outputs.long(),
            ignore_tokens=[0, 1],
        )


@register_score("core/score/rouge1")
class Rouge1Score(Score):
    def __init__(
        self,
    ):
        super().__init__()

    @classmethod
    @add_default_section_for_init("core/score/rouge1")
    def from_core_configure(cls, config, **kwargs):
        pass

    def forward(
        self,
        outputs: GenerationOutputs,
        targets: GenerationTargets,
    ):
        if isinstance(outputs, GenerationOutputs):
            outputs = outputs.sequences
        if isinstance(targets, GenerationTargets):
            targets = targets.refs

        return rouge1_score(
            targets.long(),
            outputs.long(),
            ignore_tokens=[0, 1],
        )["f1"]


@register_score("core/score/rouge2")
class Rouge2Score(Score):
    def __init__(
        self,
    ):
        super().__init__()

    @classmethod
    @add_default_section_for_init("core/score/rouge2")
    def from_core_configure(cls, config, **kwargs):
        pass

    def forward(
        self,
        outputs: GenerationOutputs,
        targets: GenerationTargets,
    ):
        if isinstance(outputs, GenerationOutputs):
            outputs = outputs.sequences
        if isinstance(targets, GenerationTargets):
            targets = targets.refs

        return rouge2_score(
            targets.long(),
            outputs.long(),
            ignore_tokens=[0, 1],
        )["f1"]


@register_score("core/score/rougel")
class RougelScore(Score):
    def __init__(
        self,
    ):
        super().__init__()

    @classmethod
    @add_default_section_for_init("core/score/rougel")
    def from_core_configure(cls, config, **kwargs):
        pass

    def forward(
        self,
        outputs: GenerationOutputs,
        targets: GenerationTargets = None,
    ):
        if isinstance(outputs, GenerationOutputs):
            outputs = outputs.sequences
        if isinstance(targets, GenerationTargets):
            targets = targets.refs

        return rougel_score(
            targets.long(),
            outputs.long(),
            ignore_tokens=[0, 1],
        )["f1"]


@register_score("core/score/loss")
class LossScore(Score):
    def __init__(
        self,
    ):
        super().__init__()

    @classmethod
    @add_default_section_for_init("core/score/loss")
    def from_core_configure(cls, config, **kwargs):
        pass

    def forward(
        self,
        outputs: LossOutputs,
        targets: ModelTargets,
    ):
        if isinstance(outputs, LossOutputs):
            loss = outputs.loss

        return -float(torch.mean(loss))


@register_score("core/score/mAP")
class MAPScore(Score):
    def __init__(
        self,
    ):
        super().__init__()

    @classmethod
    @add_default_section_for_init("core/score/mAP")
    def from_core_configure(cls, config, **kwargs):
        pass

    def forward(
        self,
        outputs: DetectionOutputs,
        targets: DetectionTargets,
    ):
        if isinstance(outputs, DetectionOutputs):
            p_bboxes = outputs.bboxes
            p_scores = outputs.scores
            p_classes = outputs.classes
        if isinstance(targets, DetectionTargets):
            gt_bboxes = targets.bboxes
            gt_classes = targets.classes
        return map_score(
            predicted_bboxes=[t.numpy() for t in p_bboxes],
            predicted_scores=[t.numpy() for t in p_scores],
            predicted_classes=[t.numpy() for t in p_classes],
            ground_truth_bboxes=[t.numpy() for t in gt_bboxes],
            ground_truth_classes=[t.numpy() for t in gt_classes],
        )


@register_score("core/score/mAP50")
class MAP50Score(Score):
    def __init__(
        self,
    ):
        super().__init__()

    @classmethod
    @add_default_section_for_init("core/score/mAP50")
    def from_core_configure(cls, config, **kwargs):
        pass

    def forward(
        self,
        outputs: DetectionOutputs,
        targets: DetectionTargets,
    ):
        if isinstance(outputs, DetectionOutputs):
            p_bboxes = outputs.bboxes
            p_scores = outputs.scores
            p_classes = outputs.classes
        if isinstance(targets, DetectionTargets):
            gt_bboxes = targets.bboxes
            gt_classes = targets.classes
        return map50_score(
            predicted_bboxes=[t.numpy() for t in p_bboxes],
            predicted_scores=[t.numpy() for t in p_scores],
            predicted_classes=[t.numpy() for t in p_classes],
            ground_truth_bboxes=[t.numpy() for t in gt_bboxes],
            ground_truth_classes=[t.numpy() for t in gt_classes],
        )



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import os
import json
import random
import logging
import torch
import torch.nn as nn
import torch.distributed as dist
import torch.nn.functional as F
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union

from transformers.models.clip.modeling_clip import (
    CLIPConfig,
    CLIPTextTransformer,
    CLIPVisionTransformer,
)
from unitorch.models import GenericModel


def _contrastive_loss(logits: torch.Tensor) -> torch.Tensor:
    return nn.functional.cross_entropy(
        logits, torch.arange(len(logits), device=logits.device)
    )


def _clip_loss(similarity: torch.Tensor) -> torch.Tensor:
    caption_loss = _contrastive_loss(similarity)
    image_loss = _contrastive_loss(similarity.T)
    return (caption_loss + image_loss) / 2.0


class AllGather(torch.autograd.Function):
    """
    Custom autograd function to perform all-gather operation on tensors.
    """

    @staticmethod
    def forward(ctx, tensor):
        """
        Forward pass of the all-gather operation.

        Args:
            ctx (torch.autograd.function.Context): Autograd function context.
            tensor (torch.Tensor): Input tensor to gather.

        Returns:
            (torch.Tensor):Gathered tensor.
        """
        ctx.rank = dist.get_rank()
        ctx.world_size = dist.get_world_size()
        y = [tensor.new(*tensor.size()) for _ in range(ctx.world_size)]
        dist.all_gather(y, tensor)
        y = torch.cat(y, 0).view(-1, *tensor.size())
        return y

    @staticmethod
    def backward(ctx, grad_output):
        """
        Backward pass of the all-gather operation.

        Args:
            ctx (torch.autograd.function.Context): Autograd function context.
            grad_output (torch.Tensor): Gradient of the output.

        Returns:
            (torch.Tensor):Gradient of the input tensor.
        """
        in_grad = torch.zeros_like(grad_output)
        in_grad.copy_(grad_output)
        # sum grad for gathered tensor
        dist.all_reduce(in_grad)
        # split
        return in_grad[ctx.rank]


class ClipForPretrain(GenericModel):
    """
    Clip model for pretraining.
    """

    def __init__(
        self,
        config_path: str,
        projection_dim: Optional[int] = 512,
        freeze_base_model: Optional[bool] = True,
        gradient_checkpointing: Optional[bool] = False,
        use_all_gather: Optional[bool] = True,
    ):
        """
        Initializes the ClipForPretrain model.

        Args:
            config_path (str): Path to the model configuration file.
            projection_dim (int, optional): Dimension of the projected embeddings. Defaults to 512.
            freeze_base_model (bool, optional): Whether to freeze the base model parameters. Defaults to True.
            gradient_checkpointing (bool, optional): Whether to use gradient checkpointing. Defaults to False.
            use_all_gather (bool, optional): Whether to use all-gather operation. Defaults to True.
        """
        super().__init__()

        config = CLIPConfig.from_json_file(config_path)
        text_config = config.text_config
        vision_config = config.vision_config
        text_config.gradient_checkpointing = gradient_checkpointing
        vision_config.gradient_checkpointing = gradient_checkpointing

        self.projection_dim = projection_dim
        self.use_all_gather = use_all_gather

        self.text_embed_dim = text_config.hidden_size
        self.vision_embed_dim = vision_config.hidden_size

        self.text_model = CLIPTextTransformer(text_config)
        self.vision_model = CLIPVisionTransformer(vision_config)

        self.visual_projection = nn.Linear(
            self.vision_embed_dim,
            self.projection_dim,
            bias=False,
        )
        self.text_projection = nn.Linear(
            self.text_embed_dim,
            self.projection_dim,
            bias=False,
        )
        self.logit_scale = nn.Parameter(torch.ones([]) * config.logit_scale_init_value)

        self.init_weights()

        if freeze_base_model:
            for p in self.text_model.parameters():
                p.requires_grad = False

            for p in self.vision_model.parameters():
                p.requires_grad = False

        self.text_model.encoder.gradient_checkpointing = gradient_checkpointing
        self.vision_model.encoder.gradient_checkpointing = gradient_checkpointing

    def _all_gather(self, input):
        """
        Perform all-gather operation on the input tensor.

        Args:
            input (torch.Tensor): Input tensor to gather.

        Returns:
            (torch.Tensor):Gathered tensor.
        """
        output = AllGather.apply(input)
        output = output.view(-1, *(output.shape[2:]))
        return output

    def forward(
        self,
        input_ids: torch.Tensor,
        pixel_values: torch.Tensor,
        attention_mask: torch.Tensor,
        position_ids: torch.Tensor,
    ):
        """
        Forward pass of the Clip model.

        Args:
            input_ids (torch.Tensor, optional): Input text token IDs. Defaults to None.
            pixel_values (torch.Tensor, optional): Input image pixel values. Defaults to None.
            attention_mask (torch.Tensor, optional): Attention mask for the input. Defaults to None.
            position_ids (torch.Tensor, optional): Position IDs for the input tokens. Defaults to None.
            output_attentions (bool, optional): Whether to output attentions. Defaults to None.
            output_hidden_states (bool, optional): Whether to output hidden states. Defaults to None.

        Returns:
            (torch.Tensor):Logits per text.
        """
        vision_outputs = self.vision_model(
            pixel_values=pixel_values,
        )

        text_outputs = self.text_model(
            input_ids=input_ids,
            attention_mask=attention_mask,
            position_ids=position_ids,
        )

        image_embeds = vision_outputs[1]
        image_embeds = self.visual_projection(image_embeds)

        text_embeds = text_outputs[1]
        text_embeds = self.text_projection(text_embeds)

        # normalized features
        image_embeds = image_embeds / image_embeds.norm(dim=-1, keepdim=True)
        text_embeds = text_embeds / text_embeds.norm(dim=-1, keepdim=True)

        logit_scale = self.logit_scale.exp()
        if self.use_all_gather and dist.is_initialized():
            text_embeds = self._all_gather(text_embeds)
            image_embeds = self._all_gather(image_embeds)
        logits_per_text = torch.matmul(text_embeds, image_embeds.t()) * logit_scale
        return _clip_loss(logits_per_text)


class ClipForClassification(GenericModel):
    def __init__(
        self,
        config_path: str,
        projection_dim: Optional[int] = 512,
        num_classes: Optional[int] = 1,
        freeze_base_model: Optional[bool] = True,
        gradient_checkpointing: Optional[bool] = False,
    ):
        """
        Clip model for classification.

        Args:
            config_path (str): Config file path to Clip model.
            projection_dim (int): Dimension for image/text output embedding.
            num_classes (int): Number of classes for classification.
            freeze_base_model (bool): Whether to freeze the base model.
            gradient_checkpointing (Optional[bool]): Whether to enable gradient_checkpointing.
        """
        super().__init__()
        config = CLIPConfig.from_json_file(config_path)
        text_config = config.text_config
        vision_config = config.vision_config
        text_config.gradient_checkpointing = gradient_checkpointing
        vision_config.gradient_checkpointing = gradient_checkpointing

        self.projection_dim = projection_dim

        self.text_embed_dim = text_config.hidden_size
        self.vision_embed_dim = vision_config.hidden_size

        self.text_model = CLIPTextTransformer(text_config)
        self.vision_model = CLIPVisionTransformer(vision_config)

        self.visual_projection = nn.Linear(
            self.vision_embed_dim,
            self.projection_dim,
            bias=False,
        )
        self.text_projection = nn.Linear(
            self.text_embed_dim,
            self.projection_dim,
            bias=False,
        )

        self.classifier = nn.Linear(self.projection_dim * 2, num_classes)

        self.init_weights()

        if freeze_base_model:
            for p in self.text_model.parameters():
                p.requires_grad = False

            for p in self.vision_model.parameters():
                p.requires_grad = False

        self.text_model.encoder.gradient_checkpointing = gradient_checkpointing
        self.vision_model.encoder.gradient_checkpointing = gradient_checkpointing

    def forward(
        self,
        input_ids: torch.Tensor,
        pixel_values: torch.Tensor,
        attention_mask: torch.Tensor,
        position_ids: torch.Tensor,
    ):
        """
        Forward pass of the Clip model for classification.

        Args:
            input_ids (tensor): Tokens of text.
            pixel_values (tensor): Pixels of image.
            attention_mask (tensor): Attention mask of tokens.
            position_ids (tensor): Position IDs.
            output_attentions (bool): Whether to output attentions.
            output_hidden_states (bool): Whether to output hidden states.

        Returns:
            tensor: Output tensor from the classifier.
        """
        vision_outputs = self.vision_model(
            pixel_values=pixel_values,
        )

        text_outputs = self.text_model(
            input_ids=input_ids,
            attention_mask=attention_mask,
            position_ids=position_ids,
        )

        image_embeds = vision_outputs[1]
        image_embeds = self.visual_projection(image_embeds)

        text_embeds = text_outputs[1]
        text_embeds = self.text_projection(text_embeds)

        return self.classifier(F.relu(torch.cat([image_embeds, text_embeds], axis=1)))


class ClipForTextClassification(GenericModel):
    def __init__(
        self,
        config_path: str,
        projection_dim: Optional[int] = 512,
        num_classes: Optional[int] = 1,
        freeze_base_model: Optional[bool] = True,
        gradient_checkpointing: Optional[bool] = False,
    ):
        """
        Initializes the Clip model for text classification.

        Args:
            config_path (str): The path to the CLIP configuration file.
            projection_dim (int, optional): The dimension of the projection layer. Defaults to 512.
            num_classes (int, optional): The number of classes for classification. Defaults to 1.
            freeze_base_model (bool, optional): Whether to freeze the base model parameters. Defaults to True.
            gradient_checkpointing (bool, optional): Whether to use gradient checkpointing. Defaults to False.
        """
        super().__init__()
        config = CLIPConfig.from_json_file(config_path)
        text_config = config.text_config
        text_config.gradient_checkpointing = gradient_checkpointing

        self.projection_dim = projection_dim
        self.text_embed_dim = text_config.hidden_size

        self.text_model = CLIPTextTransformer(text_config)

        self.text_projection = nn.Linear(
            self.text_embed_dim,
            self.projection_dim,
            bias=False,
        )

        self.classifier = nn.Linear(self.projection_dim, num_classes)

        self.init_weights()

        if freeze_base_model:
            for p in self.text_model.parameters():
                p.requires_grad = False

        self.text_model.encoder.gradient_checkpointing = gradient_checkpointing

    def forward(
        self,
        input_ids: torch.Tensor,
        attention_mask: torch.Tensor,
        position_ids: torch.Tensor,
    ):
        """
        Forward pass of the Clip model for text classification.

        Args:
            input_ids (torch.Tensor): The input token IDs.
            attention_mask (torch.Tensor): The attention mask.
            position_ids (torch.Tensor): The position IDs.

        Returns:
            (torch.Tensor):The output logits.
        """
        text_outputs = self.text_model(
            input_ids=input_ids,
            attention_mask=attention_mask,
            position_ids=position_ids,
        )
        text_embeds = text_outputs[1]
        text_embeds = self.text_projection(text_embeds)

        # normalized features
        # text_embeds = text_embeds / text_embeds.norm(dim=-1, keepdim=True)

        return self.classifier(F.relu(text_embeds))


class ClipForImageClassification(GenericModel):
    def __init__(
        self,
        config_path: str,
        projection_dim: Optional[int] = 512,
        num_classes: Optional[int] = 1,
        freeze_base_model: Optional[bool] = True,
        gradient_checkpointing: Optional[bool] = False,
    ):
        """
        Initializes the Clip model for image classification.

        Args:
            config_path (str): The path to the CLIP configuration file.
            projection_dim (int, optional): The dimension of the projection layer. Defaults to 512.
            num_classes (int, optional): The number of classes for classification. Defaults to 1.
            freeze_base_model (bool, optional): Whether to freeze the base model parameters. Defaults to True.
            gradient_checkpointing (bool, optional): Whether to use gradient checkpointing. Defaults to False.
        """
        super().__init__()
        config = CLIPConfig.from_json_file(config_path)
        vision_config = config.vision_config
        vision_config.gradient_checkpointing = gradient_checkpointing

        self.projection_dim = projection_dim
        self.vision_embed_dim = vision_config.hidden_size
        self.vision_model = CLIPVisionTransformer(vision_config)
        self.visual_projection = nn.Linear(
            self.vision_embed_dim,
            self.projection_dim,
            bias=False,
        )
        self.classifier = nn.Linear(self.projection_dim, num_classes)
        self.init_weights()

        if freeze_base_model:
            for p in self.vision_model.parameters():
                p.requires_grad = False

        self.vision_model.encoder.gradient_checkpointing = gradient_checkpointing

    def forward(
        self,
        pixel_values: torch.Tensor,
    ):
        """
        Forward pass of the Clip model for image classification.

        Args:
            pixel_values (torch.Tensor): The input pixel values.

        Returns:
            (torch.Tensor):The output logits.
        """
        vision_outputs = self.vision_model(
            pixel_values=pixel_values,
        )

        image_embeds = vision_outputs[1]
        image_embeds = self.visual_projection(image_embeds)

        # normalized features
        # image_embeds = image_embeds / image_embeds.norm(dim=-1, keepdim=True)

        return self.classifier(F.relu(image_embeds))



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

from unitorch.cli.pipelines.peft.bloom import BloomLoraForGenerationPipeline
from unitorch.cli.pipelines.peft.mistral import MistralLoraForGenerationPipeline
